# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This files contains IO operations on Feature related data.
Features are processed data linked to one or multiple record data.
For example Keypoints are features related to 1 RecordCamera (image).
And Matches are features related to a pair of RecordCamera (2 images).
"""

import numpy as np
import os
import os.path as path
from typing import Tuple, Any, Dict, Type, Optional, Union, Iterable, List

import kapture
from kapture.io.records import RECORD_DATA_DIRNAME
from kapture.utils.paths import path_secure, populate_files_in_dirpath


# Binary data File IO ##################################################################################################
def array_from_file(filepath: str, dtype: Type, dsize: int) -> np.array:
    """
    Reads a binary file of given type(dtype) and columns (dsize) into a numpy array.

    :param filepath: input path to binary file
    :param dtype: data type (eg. float)
    :param dsize: number of data (of type) per feature (eg. 2 for a simple keypoint)
    :return: numpy array
    """
    if not isinstance(dtype, type):
        raise TypeError('expect type as dtype.')
    if not isinstance(dsize, int) or dsize <= 0:
        raise TypeError('expect positive int as dsize.')

    with open(filepath, 'rb') as file:
        image_features = np.fromfile(file, dtype=dtype)
    image_features = image_features.reshape((-1, dsize))
    return image_features


def array_to_file(filepath: str, image_keypoints: np.array) -> None:
    """
    Writes the numpy array into a binary file.

    :param filepath:
    :param image_keypoints:
    """
    os.makedirs(path.dirname(filepath), exist_ok=True)
    with open(filepath, 'wb') as f:
        image_keypoints.tofile(f, sep='')


# Feature files related functions ######################################################################################
FEATURES_DATA_DIRNAMES = {
    kapture.Keypoints: path.join('reconstruction', 'keypoints'),
    kapture.Descriptors: path.join('reconstruction', 'descriptors'),
    kapture.GlobalFeatures: path.join('reconstruction', 'global_features'),
    kapture.Matches: path.join('reconstruction', 'matches'),
    kapture.RecordsBase: RECORD_DATA_DIRNAME,  # prefer kapture.RecordsCamera, Lidar, ...
}

FEATURE_FILE_EXTENSION = {
    kapture.Keypoints: '.kpt',
    kapture.Descriptors: '.desc',
    kapture.GlobalFeatures: '.gfeat',
    kapture.Matches: '.matches',
}

FEATURE_PAIR_PATH_SEPARATOR = {
    kapture.Matches: '.overlapping'
}


# get file path for binary files in kapture ############################################################################
def get_features_fullpath(
        data_type: Any,
        kapture_dirpath: str = '',
        image_filename: Optional[str] = None) -> str:
    """
    Returns full path to subdirectory containing the binary files of the given type.
            Optionally, can give a the image file name, and add the feature file name (with proper extension).

    :param data_type:
    :param kapture_dirpath: input path to kapture directory.
    :param image_filename: optional input image filename (id).
    :return: Feature full path
    """
    subdir = FEATURES_DATA_DIRNAMES[data_type]
    feature_filename = image_filename + FEATURE_FILE_EXTENSION[data_type] if image_filename else ''
    return path_secure(path.join(kapture_dirpath, subdir, feature_filename))


def features_to_filepaths(
        kapture_data: Union[kapture.RecordsCamera,
                            kapture.Keypoints,
                            kapture.Descriptors,
                            kapture.GlobalFeatures],
        kapture_dirpath: str = '') -> Dict[str, str]:
    """
    Returns a dict mapping image_id to path to feature file.
        Eg.{image_id: full/path/to/feature.jpg.kpt}

    :param kapture_data: input kapture data.
    :param kapture_dirpath: input path to kapture directory.
    :return: image id to image file path dictionary
    """
    data_type = type(kapture_data)
    return {
        image_filename: get_features_fullpath(data_type, kapture_dirpath, image_filename)
        for image_filename in kapture_data
    }


def image_ids_from_feature_dirpath(
        kapture_type: Type[Union[kapture.Keypoints,
                                 kapture.Descriptors,
                                 kapture.GlobalFeatures,
                                 kapture.Matches]],
        kapture_dirpath: str = '') -> Dict[str, str]:
    """
    Populate feature files and returns their corresponding image_filename.

    :param kapture_type:
    :param kapture_dirpath:
    :return: image file path
    """
    feature_dirpath = get_features_fullpath(kapture_type, kapture_dirpath)
    # filter only files with proper extension and
    feature_filenames = populate_files_in_dirpath(feature_dirpath, FEATURE_FILE_EXTENSION[kapture_type])
    # remove file extensions to retrieve image name.
    image_filenames = (feature_filename[:-len(FEATURE_FILE_EXTENSION[kapture_type])]
                       for feature_filename in feature_filenames)
    return image_filenames


def features_check_dir(
        kapture_data: Union[kapture.Keypoints,
                            kapture.Descriptors,
                            kapture.GlobalFeatures,
                            kapture.Matches],
        kapture_dirpath: str) -> bool:
    """
    Makes sure all files actually exists.

    :param kapture_data:
    :param kapture_dirpath:
    :return: True only if all exist, false otherwise
    """
    file_list = features_to_filepaths(kapture_data, kapture_dirpath).values()
    all_files_exists = all(path.exists(feature_filepath) for feature_filepath in file_list)
    return all_files_exists


# image_keypoints ######################################################################################################
def image_keypoints_from_file(filepath: str, dtype: Type, dsize: int) -> np.array:
    """
    Read the image keypoints

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the image keypoints
    """
    return array_from_file(filepath, dtype, dsize)


def image_keypoints_to_file(filepath: str, image_keypoints: np.array) -> None:
    """
    Writes the image keypoints to file

    :param filepath: file path
    :param image_keypoints: image keypoints
    """
    array_to_file(filepath, image_keypoints)


def get_keypoints_fullpath(kapture_dirpath: str, image_filename: Optional[str] = None) -> str:
    """
    Computes the full path of the keypoints file

    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :return: full path of the keypoints file
    """
    return get_features_fullpath(kapture.Keypoints, kapture_dirpath, image_filename)


def keypoints_to_filepaths(keypoints: kapture.Keypoints, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes keypoints files paths

    :param keypoints: keypoints
    :param kapture_dirpath: top kapture directory path
    :return: keypoint to keypoint file dictionary
    """
    return features_to_filepaths(keypoints, kapture_dirpath)


def keypoints_check_dir(keypoints: kapture.Keypoints, kapture_dirpath: str) -> bool:
    """
    Checks that all keypoints file exist.

    :param keypoints: keypoints
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(keypoints, kapture_dirpath)


# image_descriptors ####################################################################################################
def image_descriptors_from_file(filepath: str, dtype: Type, dsize: int) -> np.array:
    """
    Read the image descriptors

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the image descriptors
    """
    return array_from_file(filepath, dtype, dsize)


def image_descriptors_to_file(filepath: str, image_descriptors: np.array) -> None:
    """
    Writes the image descriptors to file

    :param filepath: file path
    :param image_descriptors: image descriptors
    """
    array_to_file(filepath, image_descriptors)


def get_descriptors_fullpath(kapture_dirpath: str, image_filename: Optional[str] = None) -> str:
    """
    Computes the full path of the descriptors file

    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :return: full path of the descriptors file
    """
    return get_features_fullpath(kapture.Descriptors, kapture_dirpath, image_filename)


def descriptors_to_filepaths(descriptors: kapture.Descriptors, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes descriptors files paths

    :param descriptors: descriptors
    :param kapture_dirpath: top kapture directory path
    :return: descriptors to descriptors file dictionary
    """
    return features_to_filepaths(descriptors, kapture_dirpath)


def descriptors_check_dir(descriptors: kapture.Descriptors, kapture_dirpath: str) -> bool:
    """
    Checks that all descriptors file exist.

    :param descriptors: descriptors
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(descriptors, kapture_dirpath)


# global_features ######################################################################################################
def image_global_features_from_file(filepath: str, dtype: Type, dsize: int) -> np.array:
    """
    Read the image global features

    :param filepath: path to the file
    :param dtype: data type
    :param dsize: number of data per keypoint
    :return: the global features
    """
    return array_from_file(filepath, dtype, dsize)


def image_global_features_to_file(filepath: str, image_global_descriptor: np.array) -> None:
    """
    Writes the image global features to file

    :param filepath: file path
    :param image_global_descriptor: image global features
    """
    array_to_file(filepath, image_global_descriptor)


def get_global_features_fullpath(kapture_dirpath: str, image_filename: Optional[str] = None) -> str:
    """
    Computes the full path of the global features file

    :param kapture_dirpath: top kapture directory path
    :param image_filename: image file name
    :return: full path of the global features file
    """
    return get_features_fullpath(kapture.GlobalFeatures, kapture_dirpath, image_filename)


def global_features_to_filepaths(global_features: kapture.GlobalFeatures, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes global features files paths

    :param global_features: global features
    :param kapture_dirpath: top kapture directory path
    :return: global features to global features file dictionary
    """
    return features_to_filepaths(global_features, kapture_dirpath)


def global_features_check_dir(global_features: kapture.GlobalFeatures, kapture_dirpath: str) -> bool:
    """
    Checks that all keypoints file exist.

    :param global_features: global features
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    return features_check_dir(global_features, kapture_dirpath)


# matches ##############################################################################################################
def image_matches_from_file(filepath: str) -> np.array:
    """
    Read the image matches
     [keypoint_id_image1, keypoint_id_image2, score]

    :param filepath: path to the file
    :return: matches
    """
    return array_from_file(filepath, dtype=np.float64, dsize=3)


def image_matches_to_file(filepath: str, image_matches: np.array) -> None:
    """
    Writes the image matches to file

    :param filepath: file path
    :param image_matches: image matches
    """

    assert image_matches.dtype == np.float64
    assert image_matches.shape[1] == 3
    array_to_file(filepath, image_matches)


def get_matches_fullpath(image_filename_pair: Optional[Tuple[str, str]] = None, kapture_dirpath: str = '') -> str:
    """
    Computes the full path of the matches file between two images

    :param kapture_dirpath: top kapture directory path
    :param image_filename_pair: image file names
    :return: full path of the matches file
    """

    if image_filename_pair is not None:
        filename = path.join(image_filename_pair[0] + FEATURE_PAIR_PATH_SEPARATOR[kapture.Matches],
                             image_filename_pair[1])
    else:
        filename = None
    return get_features_fullpath(kapture.Matches, kapture_dirpath, filename)


def matches_to_filepaths(matches: kapture.Matches, kapture_dirpath: str = '') -> Dict[Tuple[str, str], str]:
    """
    Computes matches files paths

    :param matches: matches
    :param kapture_dirpath: top kapture directory path
    :return: file to file matches to matches file dictionary
    """
    return {
        image_filename_pair: get_matches_fullpath(image_filename_pair, kapture_dirpath)
        for image_filename_pair in matches
    }


def matching_pairs_from_dirpath(kapture_dirpath: str) -> Iterable[Tuple[str, str]]:
    """
    Read and build Matches from kapture directory tree.
    """
    matches_dirpath = get_matches_fullpath(None, kapture_dirpath)
    # list all files there is
    # filter only match files (the ones endings with .matches)
    matches_filenames = populate_files_in_dirpath(matches_dirpath, FEATURE_FILE_EXTENSION[kapture.Matches])

    # remove the extensions and cut
    matching_pairs = (
        path_secure(matches_filename)[:-len(FEATURE_FILE_EXTENSION[kapture.Matches])
                                      ].split(FEATURE_PAIR_PATH_SEPARATOR[kapture.Matches] + '/')
        for matches_filename in matches_filenames
    )
    matching_pairs = ((matches[0], matches[1]) for matches in matching_pairs if len(matches) == 2)
    return matching_pairs


def matches_check_dir(matches: kapture.Matches, kapture_dirpath: str) -> bool:
    """
    Checks that all matches file exist.

    :param matches: matches
    :param kapture_dirpath: top kapture directory path
    :return: True if they all exist, false otherwise.
    """
    file_list = (get_matches_fullpath(pair, kapture_dirpath) for pair in matches)
    all_files_exists = all(path.exists(filepath) for filepath in file_list)
    return all_files_exists



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Evaluation with kapture objects
"""

import math
from typing import Union, List, Tuple, Set

import kapture
from .pose_operations import world_pose_transform_distance


def evaluate_error_absolute(poses_to_test: List[Tuple[str, kapture.PoseTransform]],
                            poses_ground_truth: List[Tuple[str, kapture.PoseTransform]]
                            ) -> List[Tuple[str, float, float]]:
    """
    Evaluate the absolute error for poses to a ground truth.

    :param poses_to_test: poses to test
    :param poses_ground_truth: reference poses
    :return: list of error evaluation
    """
    poses_ground_truth_as_dict = {name: pose for name, pose in poses_ground_truth}
    result = [(name,) + world_pose_transform_distance(pose, poses_ground_truth_as_dict[name])
              for (name, pose) in poses_to_test]
    return result


def get_poses(k_data: kapture.Kapture,
              image_set: Union[Set[str], List[str]]) -> List[Tuple[str, kapture.PoseTransform]]:
    """
    Computes the poses for a set of images within a kapture.

    :param k_data: the kapture
    :param image_set: set of image names
    :return: list of (image name,pose)
    """
    assert k_data.trajectories is not None
    if isinstance(image_set, list):
        image_set = set(image_set)
    assert isinstance(image_set, set)
    assert isinstance(k_data, kapture.Kapture)

    # apply rigs to trajectories
    if k_data.rigs is not None:
        trajectories = kapture.rigs_remove(k_data.trajectories, k_data.rigs)
    else:
        trajectories = k_data.trajectories

    poses = []
    for timestamp, device_id, filename in kapture.flatten(k_data.records_camera, is_sorted=True):
        if filename in image_set and (timestamp, device_id) in trajectories:
            pose = trajectories[(timestamp, device_id)]
            poses.append((filename, pose))
    return poses


# def get_imagename_intersection(kdata_list: List[kapture.Kapture]) -> Set[str]:
#     def get_image_set(k_data: kapture.Kapture):
#         return set(image_name for _, _, image_name in kapture.flatten(k_data.records_camera)) \
#             if k_data.records_camera is not None else set()
#     assert isinstance(kdata_list, list)
#     assert len(kdata_list) > 0
#     return set.intersection(*[get_image_set(k_data) for k_data in kdata_list])


def evaluate(k_data: kapture.Kapture,
             k_data_gt: kapture.Kapture,
             image_set: Union[Set[str], List[str]]) -> List[Tuple[str, float, float]]:
    """
    Evaluate the pose found for images in a kapture with a reference kapture.

    :param k_data: the kapture to test
    :param k_data_gt: the reference kapture
    :param image_set: list of image names
    :return: list of image pose evaluation
    """
    if isinstance(image_set, list):
        image_set = set(image_set)
    assert isinstance(image_set, set)
    assert(len(image_set) > 0)
    assert isinstance(k_data, kapture.Kapture)
    assert isinstance(k_data_gt, kapture.Kapture)

    poses_to_test = get_poses(k_data, image_set)
    poses_gt = get_poses(k_data_gt, image_set)

    evaluated = evaluate_error_absolute(poses_to_test, poses_gt)
    localized_images = {name for name, position_error, rotation_error in evaluated}
    missing_images = [name for name in image_set if name not in localized_images]
    for name in missing_images:
        evaluated.append((name, math.nan, math.nan))
    return sorted(evaluated)


def fill_bins(results: List[Tuple[str, float, float]],
              bins: List[Tuple[float, float]]
              ) -> List[Tuple[float, float, int]]:
    """
    Fill a bin with the number of images within position thresholds.

    :param results: list of error evaluation (image name, translation error, rotation error)
    :param bins: list of translation and rotation thresholds
    :return: number of images in every pair of (translation,rotation) error
    """
    assert isinstance(results, list)
    assert isinstance(bins, list)

    all_positions = [(translation_error, rotation_error) for name, translation_error, rotation_error in results]
    filled_bins = []
    for a_bin in bins:
        trans_threshold = a_bin[0]
        rot_threshold = a_bin[1]
        number_of_images_in_bin = 0
        for translation_error, rotation_error in all_positions:
            if (math.isnan(rot_threshold) or rot_threshold < 0) and translation_error <= trans_threshold:
                number_of_images_in_bin += 1
            elif translation_error <= trans_threshold and rotation_error <= rot_threshold:
                number_of_images_in_bin += 1
        filled_bins.append((trans_threshold, rot_threshold, number_of_images_in_bin))
    return filled_bins



#!/usr/bin/env python3
# Copyright (c) 2019, Naver Labs Europe
# This script imports data from a COLMAP database and/or reconstruction files

import argparse
import logging
import os
import os.path as path
import json
import numpy as np
import quaternion
import pickle
import gzip
from tqdm import tqdm
from typing import Dict, Optional
import path_to_kapture  # enables import kapture
import kapture
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.converter.opensfm.import_opensfm import import_opensfm


logger = logging.getLogger('opensfm')


def import_opensfm_command_line():
    """
    Imports openSfM to kapture.
    """
    parser = argparse.ArgumentParser(description='Imports openSfM to kapture')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    ####################################################################################################################
    parser.add_argument('-i', '--input', '--opensfm', required=True,
                        help='path to OpenSfM root directory.')
    parser.add_argument('-o', '-k', '--output', '--kapture', required=True,
                        help='output directory where to save kapture files.')
    parser.add_argument('--transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete kapture if already exists.')
    args = parser.parse_args()
    ####################################################################################################################
    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    logger.debug(f'\\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()
        if k is not 'command'))

    args.input = path.normpath(path.abspath(args.input))
    args.output = path.normpath(path.abspath(args.output))

    import_opensfm(
        opensfm_rootdir=args.input,
        kapture_rootdir=args.output,
        force_overwrite_existing=args.force,
        images_import_method=args.transfer
    )


if __name__ == '__main__':
    import_opensfm_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from typing import Union, Dict, List, Tuple, TypeVar, Optional
from .flatten import flatten


class RecordWifi:
    """
    RSSI, FREQ, SCANTIME, VISIBLENAME
    """

    def __init__(self,
                 rssi: int,
                 freq: int,
                 scan_time: int,
                 visible_name: str):
        # enforce type
        self.rssi = int(rssi)
        self.freq = int(freq)
        self.scan_time = int(scan_time)
        self.visible_name = str(visible_name)

    def as_list(self) -> List[str]:
        """
        :return: Wifi records as list of strings
        """
        values = [self.rssi, self.freq,
                  self.scan_time, self.visible_name]
        return [str(v) for v in values]

    def __repr__(self) -> str:
        return ', '.join(self.as_list())

    def __hash__(self):
        return hash(tuple(self.as_list()))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.as_list() == other.as_list()


class RecordGnss:
    """
    X, Y, Z, UTC, DOP,
    """
    def __init__(self,
                 x: float, y: float, z: float,
                 utc: int,
                 dop: float):
        # enforce type
        self.x, self.y, self.z = float(x), float(y), float(z)
        self.utc = int(utc)
        self.dop = float(dop)

    def as_list(self) -> List[str]:
        """
        :return: GNSS records as list of strings
        """
        values = [self.x, self.y, self.z, self.utc, self.dop]
        return [str(v) for v in values]

    def __repr__(self) -> str:
        return ', '.join(self.as_list())

    def __hash__(self):
        return hash(tuple(self.as_list()))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.as_list() == other.as_list()


########################################################################################################################
T = TypeVar('T')  # Declare generic type variable


class RecordsBase(Dict[int, Dict[str, T]]):
    """
    brief: Records
            records[timestamp][sensor_id] = <DataRecords>
            or
            records[(timestamp, sensor_id)] = <DataRecords>
    """

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, T], T]):
        """
        Inserts/changes a record.

        :param key: can be timestamp or (timestamp, sensor_id)
        :param value: can be respectively a dict{sensor_id: DataRecords} or DataRecords
        :return:
        """
        # enforce type checking
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            self.setdefault(timestamp, {})[device_id] = value
        elif isinstance(key, int):
            # key is a timestamp
            timestamp = key
            if not isinstance(value, dict):
                raise TypeError('invalid value for data')
            if not all(isinstance(k, str) for k in value.keys()):
                raise TypeError('invalid device_id')
            super(RecordsBase, self).__setitem__(timestamp, value)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __getitem__(self, key: Union[int, Tuple[int, str]]) -> Union[Dict[str, T], T]:
        """
        Returns a single record for a (timestamp, sensor) or all records for a timestamp.

        :param key: can be timestamp or (timestamp, sensor_id)
        :return: return respectively a dict{sensor_id: DataRecords} or DataRecords
        """
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(RecordsBase, self).__getitem__(timestamp)[device_id]
        elif isinstance(key, int):
            # key is a timestamp
            return super(RecordsBase, self).__getitem__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def key_pairs(self) -> List[Tuple[int, str]]:
        """
        Returns a list of (timestamp, device_id) contained in records.
        Those pairs can be used to access a single record data.

        :return: list of (timestamp, device_id)
        """
        return [
            (timestamp, sensor_id)
            for timestamp, sensors in self.items()
            for sensor_id in sensors.keys()
        ]

    def __contains__(self, key: Union[int, Tuple[int, str]]):
        if isinstance(key, tuple):
            # key is a pair of (timestamp, device_id)
            timestamp = key[0]
            device_id = key[1]
            if not isinstance(timestamp, int):
                raise TypeError('invalid timestamp')
            if not isinstance(device_id, str):
                raise TypeError('invalid device_id')
            return super(RecordsBase, self).__contains__(timestamp) and device_id in self[timestamp]
        elif isinstance(key, int):
            return super(RecordsBase, self).__contains__(key)
        else:
            raise TypeError('key must be Union[int, Tuple[int, str]]')

    def __repr__(self) -> str:
        # [timestamp, sensor_id] = str
        lines = [f'[ {timestamp:010}, {sensor_id:5}] = {data}'
                 for timestamp, sensors in self.items()
                 for sensor_id, data in sensors.items()]
        return '\n'.join(lines)


class RecordsCamera(RecordsBase[str]):
    """
    Camera records
    """

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, str], str]):
        """ see RecordsBase.__setitem__ """
        if isinstance(key, tuple):
            if not isinstance(value, str):
                raise TypeError('invalid data')
        elif isinstance(key, int):
            if not isinstance(value, dict):
                raise TypeError('invalid value for data')
            if not all(isinstance(v, str) for v in value.values()):
                raise TypeError('invalid data')
        super(RecordsCamera, self).__setitem__(key, value)


class RecordsLidar(RecordsBase[str]):
    """
    Lidar records
    """

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, str], str]):
        """ see RecordsBase.__setitem__ """
        if isinstance(key, tuple):
            if not isinstance(value, str):
                raise TypeError('invalid data')
        elif isinstance(key, int):
            if not isinstance(value, dict):
                raise TypeError('invalid value for data')
            if not all(isinstance(v, str) for v in value.values()):
                raise TypeError('invalid data')
        super(RecordsLidar, self).__setitem__(key, value)


class RecordsWifi(RecordsBase[Dict[str, RecordWifi]]):
    """
    brief: Records
            records[timestamp][sensor_id][bssid] = <RecordWifi>
            or
            records[(timestamp, sensor_id)][bssid] = <RecordWifi>
    """

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Dict[str, RecordWifi]):
        if isinstance(key, tuple):
            if not isinstance(value, Dict):
                raise TypeError('invalid data')
        elif isinstance(key, int):
            if not isinstance(value, dict):
                raise TypeError('invalid value for data')
            if not all(isinstance(v, dict) for v in value.values()):
                raise TypeError('invalid data')
        super(RecordsWifi, self).__setitem__(key, value)


class RecordsGnss(RecordsBase[Dict[str, RecordGnss]]):
    """
    brief: Records
            records[timestamp][sensor_id] = <RecordGnss>
            or
            records[(timestamp, sensor_id)] = <RecordGnss>
    """

    def __setitem__(self,
                    key: Union[int, Tuple[int, str]],
                    value: Union[Dict[str, RecordGnss], RecordGnss]):
        """
        will raise error is sensor_id is not valid.
        """
        if isinstance(key, tuple):
            # records[timestamp, sensor_id] = <RecordGnss>
            # check record is valid
            if not isinstance(value, RecordGnss):
                raise TypeError('invalid value for data')
        elif isinstance(key, int):
            # records[timestamp] = {sensor_id: <RecordGnss>}
            if not isinstance(value, dict):
                raise TypeError('invalid value for data')
            # check all values are valid records
            if not all(isinstance(v, RecordGnss) for v in value.values()):
                raise TypeError('invalid GNSS data')

        super(RecordsGnss, self).__setitem__(key, value)

    def __repr__(self) -> str:
        lines = []
        # [timestamp, sensor_id] = <RecordGnss>
        lines += [f'[ {timestamp:010}, {sensor_id:5}] = {location}'
                  for timestamp, sensor_id, location in flatten(self)]
        return '\n'.join(lines)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This files contains IO operations on Record related data.
"""

from enum import auto
import logging
import os
import os.path as path
import shutil
from typing import Tuple, Any, Dict, Type, Optional, Union, Iterable, List
from tqdm import tqdm

import kapture
from kapture.utils.paths import path_secure
from kapture.utils.Collections import AutoEnum
from kapture.utils.logging import getLogger

logger = getLogger()

# Records files related functions ######################################################################################
RECORD_DATA_DIRNAME = path_secure(path.join('sensors', 'records_data'))


class TransferAction(AutoEnum):
    """
    All possible operations when transferring recorded data (eg. images, Wifi, IMU, ...)
    or computed data (features, keypoints, ...).
    Transferring means importing, exporting, and more generally going from one representation
    to another representation of the data.
    """
    skip = auto()
    root_link = auto()
    copy = auto()
    move = auto()
    link_absolute = auto()
    link_relative = auto()


########################################################################################################################
# transfer files functions
def transfer_files_from_dir_link(
        source_filepath_list: Iterable[str],
        destination_filepath_list: Iterable[str],
        force_overwrite: bool = False,
        do_relative_link: bool = False
) -> None:
    """
    Transfer every files by linking given from the source list to destination list.
    The matching between source and kapture files are explicitly given.

    :param source_filepath_list: input list of source files. Uses guess_filepaths to obtains it from filenames.
    :param destination_filepath_list: input list of destination files (in kapture tree).
    :param force_overwrite: if True, overwrite destination file.
    :param do_relative_link: if True, do relative links else absolute links.
    """
    hide_progress_bar = logger.getEffectiveLevel() > logging.INFO
    for src, dst in tqdm(zip(source_filepath_list, destination_filepath_list), disable=hide_progress_bar):
        os.makedirs(path.dirname(dst), exist_ok=True)
        if force_overwrite and path.lexists(dst):
            os.remove(dst)
        try:  # on windows, symlink requires some privileges, and may crash if not
            if do_relative_link:
                src = path.relpath(src, path.dirname(dst))
            os.symlink(src, dst)
        except OSError as e:
            logger.critical('unable to create symlink on image directory, due to privilege restrictions.')
            raise e


def transfer_files_from_dir_copy(
        source_filepath_list: Iterable[str],
        destination_filepath_list: Iterable[str],
        force_overwrite: bool = False,
        delete_source: bool = False
) -> None:
    """
    Transfer every files by copying given from the source list to destination list.
    The matching between source and kapture files are explicitly given.
    If delete_source is activated, it moves files instead copying them.

    :param source_filepath_list: input list of absolute path to source files.
    :param destination_filepath_list: input list of absolute path to destination files.
    :param force_overwrite: if True, overwrite destination file.
    :param delete_source: if True, delete the imported files from source_record_dirpath.
    """
    hide_progress_bar = logger.getEffectiveLevel() > logging.INFO
    for src, dst in tqdm(zip(source_filepath_list, destination_filepath_list), disable=hide_progress_bar):
        os.makedirs(path.dirname(dst), exist_ok=True)
        if force_overwrite and path.lexists(dst):
            os.remove(dst)
        if delete_source:
            shutil.move(src, dst)
        else:
            shutil.copyfile(src, dst)


def transfer_files_from_dir(
        source_filepath_list: Iterable[str],
        destination_filepath_list: Iterable[str],
        copy_strategy: TransferAction = TransferAction.copy,
        force_overwrite: bool = False
) -> None:
    """
    Transfer files (copy or link files) from source to destination.
    The matching between source and kapture files are explicitly given.
    The actual import can be done using actual file copy (copy/move), or symlinks (absolute/relative).

    :param source_filepath_list: input list of source files.
    :param destination_filepath_list: input list of destination files (in kapture tree).
    :param copy_strategy: transfer strategy to apply.
    :param force_overwrite: if True, overwrite destination file.
    """
    if TransferAction.skip == copy_strategy:
        return

    elif TransferAction.copy == copy_strategy or TransferAction.move == copy_strategy:
        transfer_files_from_dir_copy(
            source_filepath_list=source_filepath_list,
            destination_filepath_list=destination_filepath_list,
            force_overwrite=force_overwrite,
            delete_source=(TransferAction.move == copy_strategy)
        )

    elif TransferAction.link_absolute == copy_strategy or TransferAction.link_relative == copy_strategy:
        transfer_files_from_dir_link(
            source_filepath_list=source_filepath_list,
            destination_filepath_list=destination_filepath_list,
            force_overwrite=force_overwrite,
            do_relative_link=(TransferAction.link_relative == copy_strategy)
        )

    else:
        raise ValueError(f'Unsupported copy action {copy_strategy}')


########################################################################################################################
def get_record_fullpath(
        kapture_dirpath: str = '',
        record_filename: Optional[str] = None) -> str:
    """
    Returns full path to subdirectory containing the binary files of the record type.
            Optionally, can give a the file name.
    :param kapture_dirpath: input path to kapture directory.
    :param record_filename: optional input record filename (eg. image filename).
    :return: the record full path
    """
    feature_filename = record_filename or ''
    return path_secure(path.join(kapture_dirpath, RECORD_DATA_DIRNAME, feature_filename))


def records_to_filepaths(
        records: Union[kapture.RecordsCamera, kapture.RecordsWifi, kapture.RecordsLidar, kapture.RecordsGnss],
        kapture_dirpath: str
) -> Dict[str, str]:
    """
    Computes filepaths for records.

    :param records: records
    :param kapture_dirpath: top kapture directory path
    :return: records name to records file path dictionary
    """
    return {filename: path_secure(path.join(kapture_dirpath, RECORD_DATA_DIRNAME, filename))
            for _, _, filename in kapture.flatten(records)}


def guess_filepaths_from_filenames(
        dirpath: str,
        filenames: Iterable[str]
):
    """ returns a generator that prepend the directory path to the given filenames."""
    return (path_secure(path.join(dirpath, record_filename))
            for record_filename in filenames)


def import_record_data_from_dir_link_dir(
        source_record_dirpath: str,
        kapture_dirpath: str,
        do_relative_link: bool = False
) -> None:
    """
    Imports record_data from a given directory:
    This do not actually copy any files, but make a symbolic link on the root path instead.

    :param source_record_dirpath: input path to directory where to import data from.
    :param kapture_dirpath: input root path to kapture root directory, where to import data to.
    :param do_relative_link: if True, do relative link else absolute link.
    """
    assert path.isdir(source_record_dirpath)
    kapture_record_dirpath = get_record_fullpath(kapture_dirpath)
    if do_relative_link:
        source_record_dirpath = path.relpath(source_record_dirpath, kapture_record_dirpath)
    try:  # on windows, symlink requires some privileges, and may crash if not
        os.symlink(source_record_dirpath, kapture_record_dirpath)
    except OSError as e:
        logger.critical('unable to create symlink on image directory, due to privilege restrictions.')
        raise e


def import_record_data_from_dir_auto(
        source_record_dirpath: str,
        destination_kapture_dirpath: str,
        filename_list: Iterable[str],
        copy_strategy: TransferAction = TransferAction.copy
) -> None:
    """
    Imports record_data from a given directory.
    Automatically replicate the folder hierarchy from source to kapture record directory.
    If you want to change the file organisation use import_record_data_from_dir_explicit instead.
    The actual import can be done using actual file copy (copy/move), or symlinks (absolute/relative).

    :param source_record_dirpath: input path to directory where to import data from.
    :param destination_kapture_dirpath: input root path to kapture root directory, where to import data to.
    :param filename_list: input list of filenames to import (filenames are relative path to source_record_dirpath).
                    This image list can be obtained from file walking (populate_files_in_dirpath)
                    or using an already populated kapture.RecordsData. Prefer RecordsData since it will only copy
                    required files, eg:
                    filename_list = [f for _, _, f in kapture.flatten(kapture_data.records_camera)]
    :param copy_strategy:
    """
    if copy_strategy == TransferAction.root_link:
        import_record_data_from_dir_link_dir(source_record_dirpath, destination_kapture_dirpath)
    else:
        source_filepath_list = (path.join(source_record_dirpath, record_filename)
                                for record_filename in filename_list)
        kapture_filepath_list = (get_record_fullpath(destination_kapture_dirpath, record_filename)
                                 for record_filename in filename_list)
        transfer_files_from_dir(
            source_filepath_list,
            kapture_filepath_list,
            copy_strategy
        )


# images ###############################################################################################################
def get_image_fullpath(kapture_dir_path: str, image_filename: Optional[str] = None) -> str:
    """
    Get the full path of the image file in the kapture.
     If image file name is missing, this gives the top directory of the images.

    :param kapture_dir_path: the kapture top directory
    :param image_filename: optional image file name
    :return: Image file full path
    """
    return get_record_fullpath(kapture_dir_path, image_filename)


def images_to_filepaths(images: kapture.RecordsCamera, kapture_dirpath: str) -> Dict[str, str]:
    """
    Computes filepaths for image records.

    :param images: images records
    :param kapture_dirpath: top kapture directory path
    :return: images name to images file path dictionary
    """
    return records_to_filepaths(images, kapture_dirpath)



import logging
import os
import os.path as path
import numpy as np
import quaternion
import gzip
import pickle
import json
from tqdm import tqdm
from typing import Optional, Dict
# kapture
import kapture
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.io.features import get_keypoints_fullpath, image_keypoints_from_file
from kapture.io.features import get_descriptors_fullpath, image_descriptors_from_file
from kapture.io.features import get_matches_fullpath, image_matches_from_file
from kapture.io.records import TransferAction, transfer_files_from_dir, get_record_fullpath
import kapture.io.csv as csv
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
from kapture.io.structure import delete_existing_kapture_files

logger = logging.getLogger('opensfm')

"""
opensfm_project/
├── config.yaml
├── images/
├── masks/
├── gcp_list.txt
├── exif/
├── camera_models.json
├── features/
├── matches/
├── tracks.csv
├── reconstruction.json
├── reconstruction.meshed.json
└── undistorted/
    ├── images/
    ├── masks/
    ├── tracks.csv
    ├── reconstruction.json
    └── depthmaps/
        └── merged.ply
"""

"""
reconstruction.json: [RECONSTRUCTION, ...]

RECONSTRUCTION: {
    "cameras": {
        CAMERA_ID: CAMERA,
        ...
    },
    "shots": {
        SHOT_ID: SHOT,
        ...
    },
    "points": {
        POINT_ID: POINT,
        ...
    }
}

CAMERA: {
    "projection_type": "perspective",  # Can be perspective, brown, fisheye or equirectangular
    "width": NUMBER,                   # Image width in pixels
    "height": NUMBER,                  # Image height in pixels

    # Depending on the projection type more parameters are stored.
    # These are the parameters of the perspective camera.
    "focal": NUMBER,                   # Estimated focal length
    "k1": NUMBER,                      # Estimated distortion coefficient
    "k2": NUMBER,                      # Estimated distortion coefficient
}

SHOT: {
    "camera": CAMERA_ID,
    "rotation": [X, Y, Z],      # Estimated rotation as an angle-axis vector
    "translation": [X, Y, Z],   # Estimated translation
    "gps_position": [X, Y, Z],  # GPS coordinates in the reconstruction reference frame
    "gps_dop": METERS,          # GPS accuracy in meters
    "orientation": NUMBER,      # EXIF orientation tag (can be 1, 3, 6 or 8)
    "capture_time": SECONDS     # Capture time as a UNIX timestamp
}

POINT: {
    "coordinates": [X, Y, Z],      # Estimated position of the point
    "color": [R, G, B],            # Color of the point
}
"""


def import_camera(
        opensfm_camera: Dict[str, float],
        name: Optional[str] = None
) -> kapture.Camera:
    # opensfm_camera['projection_type'] can be perspective, brown, fisheye or equirectangular
    if 'perspective' == opensfm_camera['projection_type']:
        # convert to CameraType.RADIAL [w, h, f, cx, cy, k1, k2]
        # missing principal point, just fake it at image center
        largest_side_in_pixel = float(max(opensfm_camera['width'], opensfm_camera['height']))
        camera_params = [
            # w, h:
            opensfm_camera['width'], opensfm_camera['height'],
            # f: The focal length provided by the EXIF metadata divided by the sensor width
            opensfm_camera['focal'] * largest_side_in_pixel,
            # cx, cy: no principal point, guess one at image center
            opensfm_camera['width'] / 2, opensfm_camera['height'] / 2,
            # k1, k2
            opensfm_camera.get('k1', 0.0), opensfm_camera.get('k2', 0.0),
        ]
        return kapture.Camera(
            camera_type=kapture.CameraType.RADIAL,
            camera_params=camera_params,
            name=name
        )

    else:
        raise ValueError(f'unable to convert camera of type {opensfm_camera["projection_type"]}')


def import_opensfm(
        opensfm_rootdir: str,
        kapture_rootdir: str,
        force_overwrite_existing: bool = False,
        images_import_method: TransferAction = TransferAction.copy
) -> None:
    disable_tqdm = logger.getEffectiveLevel() != logging.INFO
    # load reconstruction
    opensfm_reconstruction_filepath = path.join(opensfm_rootdir, 'reconstruction.json')
    with open(opensfm_reconstruction_filepath, 'rt') as f:
        opensfm_reconstruction = json.load(f)
    # remove the single list @ root
    opensfm_reconstruction = opensfm_reconstruction[0]

    # prepare space for output
    os.makedirs(kapture_rootdir, exist_ok=True)
    delete_existing_kapture_files(kapture_rootdir, force_erase=force_overwrite_existing)

    # import cameras
    kapture_sensors = kapture.Sensors()
    assert 'cameras' in opensfm_reconstruction
    # import cameras
    for osfm_camera_id, osfm_camera in opensfm_reconstruction['cameras'].items():
        camera = import_camera(osfm_camera, name=osfm_camera_id)
        kapture_sensors[osfm_camera_id] = camera

    # import shots
    logger.info('importing images and trajectories ...')
    kapture_images = kapture.RecordsCamera()
    kapture_trajectories = kapture.Trajectories()
    opensfm_image_dirpath = path.join(opensfm_rootdir, 'images')
    assert 'shots' in opensfm_reconstruction
    image_timestamps, image_sensors = {}, {}  # used later to retrieve the timestamp of an image.
    for timestamp, (image_filename, shot) in enumerate(opensfm_reconstruction['shots'].items()):
        sensor_id = shot['camera']
        image_timestamps[image_filename] = timestamp
        image_sensors[image_filename] = sensor_id
        # in OpenSfm, (sensor, timestamp) is not unique.
        rotation_vector = shot['rotation']
        q = quaternion.from_rotation_vector(rotation_vector)
        translation = shot['translation']
        # capture_time = shot['capture_time'] # may be invalid
        # gps_position = shot['gps_position']
        kapture_images[timestamp, sensor_id] = image_filename
        kapture_trajectories[timestamp, sensor_id] = kapture.PoseTransform(r=q, t=translation)

    # copy image files
    filename_list = [f for _, _, f in kapture.flatten(kapture_images)]
    import_record_data_from_dir_auto(
        source_record_dirpath=opensfm_image_dirpath,
        destination_kapture_dirpath=kapture_rootdir,
        filename_list=filename_list,
        copy_strategy=images_import_method)

    # gps from pre-extracted exif, in exif/image_name.jpg.exif
    kapture_gnss = None
    opensfm_exif_dirpath = path.join(opensfm_rootdir, 'exif')
    opensfm_exif_suffix = '.exif'
    if path.isdir(opensfm_exif_dirpath):
        logger.info('importing GNSS from exif ...')
        camera_ids = set(image_sensors.values())
        # add a gps sensor for each camera
        map_cam_to_gnss_sensor = {cam_id: 'GPS_' + cam_id for cam_id in camera_ids}
        for gnss_id in map_cam_to_gnss_sensor.values():
            kapture_sensors[gnss_id] = kapture.Sensor(sensor_type='gnss', sensor_params=['EPSG:4326'])
        # build epsg_code for all cameras
        kapture_gnss = kapture.RecordsGnss()
        opensfm_exif_filepath_list = (path.join(dirpath, filename)
                                      for dirpath, _, filename_list in os.walk(opensfm_exif_dirpath)
                                      for filename in filename_list
                                      if filename.endswith(opensfm_exif_suffix))
        for opensfm_exif_filepath in tqdm(opensfm_exif_filepath_list, disable=disable_tqdm):
            image_filename = path.relpath(opensfm_exif_filepath, opensfm_exif_dirpath)[:-len(opensfm_exif_suffix)]
            image_timestamp = image_timestamps[image_filename]
            image_sensor_id = image_sensors[image_filename]
            gnss_timestamp = image_timestamp
            gnss_sensor_id = map_cam_to_gnss_sensor[image_sensor_id]
            with open(opensfm_exif_filepath, 'rt') as f:
                js_root = json.load(f)
                if 'gps' not in js_root:
                    logger.warning(f'NO GPS data in "{opensfm_exif_filepath}"')
                    continue

                gps_coords = {
                    'x': js_root['gps']['longitude'],
                    'y': js_root['gps']['latitude'],
                    'z': js_root['gps'].get('altitude', 0.0),
                    'dop': js_root['gps'].get('dop', 0),
                    'utc': 0,
                }
                logger.debug(f'found GPS data for ({gnss_timestamp}, {gnss_sensor_id}) in "{opensfm_exif_filepath}"')
                kapture_gnss[gnss_timestamp, gnss_sensor_id] = kapture.RecordGnss(**gps_coords)

    # import features (keypoints + descriptors)
    kapture_keypoints = None  # kapture.Keypoints(type_name='opensfm', dsize=4, dtype=np.float64)
    kapture_descriptors = None  # kapture.Descriptors(type_name='opensfm', dsize=128, dtype=np.uint8)
    opensfm_features_dirpath = path.join(opensfm_rootdir, 'features')
    opensfm_features_suffix = '.features.npz'
    if path.isdir(opensfm_features_dirpath):
        logger.info('importing keypoints and descriptors ...')
        opensfm_features_file_list = (path.join(dp, fn)
                                      for dp, _, fs in os.walk(opensfm_features_dirpath) for fn in fs)
        opensfm_features_file_list = (filepath
                                      for filepath in opensfm_features_file_list
                                      if filepath.endswith(opensfm_features_suffix))
        for opensfm_feature_filename in tqdm(opensfm_features_file_list, disable=disable_tqdm):
            image_filename = path.relpath(opensfm_feature_filename, opensfm_features_dirpath)[
                             :-len(opensfm_features_suffix)]
            opensfm_image_features = np.load(opensfm_feature_filename)
            opensfm_image_keypoints = opensfm_image_features['points']
            opensfm_image_descriptors = opensfm_image_features['descriptors']
            logger.debug(f'parsing keypoints and descriptors in {opensfm_feature_filename}')
            if kapture_keypoints is None:
                # print(type(opensfm_image_keypoints.dtype))
                # HAHOG = Hessian Affine feature point detector + HOG descriptor
                kapture_keypoints = kapture.Keypoints(
                    type_name='HessianAffine',
                    dsize=opensfm_image_keypoints.shape[1],
                    dtype=opensfm_image_keypoints.dtype)
            if kapture_descriptors is None:
                kapture_descriptors = kapture.Descriptors(
                    type_name='HOG',
                    dsize=opensfm_image_descriptors.shape[1],
                    dtype=opensfm_image_descriptors.dtype)

            # convert keypoints file
            keypoint_filpath = kapture.io.features.get_features_fullpath(
                data_type=kapture.Keypoints, kapture_dirpath=kapture_rootdir, image_filename=image_filename)
            kapture.io.features.image_keypoints_to_file(
                filepath=keypoint_filpath, image_keypoints=opensfm_image_keypoints)
            # register the file
            kapture_keypoints.add(image_filename)

            # convert descriptors file
            descriptor_filpath = kapture.io.features.get_features_fullpath(
                data_type=kapture.Descriptors, kapture_dirpath=kapture_rootdir, image_filename=image_filename)
            kapture.io.features.image_descriptors_to_file(
                filepath=descriptor_filpath, image_descriptors=opensfm_image_descriptors)
            # register the file
            kapture_descriptors.add(image_filename)

    # import matches
    kapture_matches = kapture.Matches()
    opensfm_matches_suffix = '_matches.pkl.gz'
    opensfm_matches_dirpath = path.join(opensfm_rootdir, 'matches')
    if path.isdir(opensfm_matches_dirpath):
        logger.info('importing matches ...')
        opensfm_matches_file_list = (path.join(dp, fn)
                                     for dp, _, fs in os.walk(opensfm_matches_dirpath) for fn in fs)
        opensfm_matches_file_list = (filepath
                                     for filepath in opensfm_matches_file_list
                                     if filepath.endswith(opensfm_matches_suffix))

        for opensfm_matches_filename in tqdm(opensfm_matches_file_list, disable=disable_tqdm):
            image_filename_1 = path.relpath(opensfm_matches_filename, opensfm_matches_dirpath)[
                               :-len(opensfm_matches_suffix)]
            logger.debug(f'parsing mathes in {image_filename_1}')
            with gzip.open(opensfm_matches_filename, 'rb') as f:
                opensfm_matches = pickle.load(f)
                for image_filename_2, opensfm_image_matches in opensfm_matches.items():
                    image_pair = (image_filename_1, image_filename_2)
                    # register the pair to kapture
                    kapture_matches.add(*image_pair)
                    # convert the bin file to kapture
                    kapture_matches_filepath = kapture.io.features.get_matches_fullpath(
                        image_filename_pair=image_pair,
                        kapture_dirpath=kapture_rootdir)
                    kapture_image_matches = np.hstack([
                        opensfm_image_matches.astype(np.float64),
                        # no macthes scoring = assume all to one
                        np.ones(shape=(opensfm_image_matches.shape[0], 1), dtype=np.float64)])
                    kapture.io.features.image_matches_to_file(kapture_matches_filepath, kapture_image_matches)

    # import 3-D points
    if 'points' in opensfm_reconstruction:
        logger.info('importing points 3-D')
        opensfm_points = opensfm_reconstruction['points']
        points_data = []
        for point_id in sorted(opensfm_points):
            point_data = opensfm_points[point_id]
            point_data = point_data['coordinates'] + point_data['color']
            points_data.append(point_data)
        kapture_points = kapture.Points3d(points_data)
    else:
        kapture_points = None

    # saving kapture csv files
    logger.info('saving kapture files')
    kapture_data = kapture.Kapture(
        sensors=kapture_sensors,
        records_camera=kapture_images,
        records_gnss=kapture_gnss,
        trajectories=kapture_trajectories,
        keypoints=kapture_keypoints,
        descriptors=kapture_descriptors,
        matches=kapture_matches,
        points3d=kapture_points
    )
    kapture.io.csv.kapture_to_dir(dirpath=kapture_rootdir, kapture_data=kapture_data)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import argparse
import logging
import os
import pathlib
from typing import List, Tuple, Union
import numpy as np
from tqdm import tqdm

import path_to_kapture
import kapture
import kapture.utils.logging
from kapture.io.csv import kapture_from_dir, table_to_file
from kapture.io.csv import ImageFeatureConfig
from kapture.io.features import image_global_features_from_file, global_features_to_filepaths

logger = logging.getLogger('compute_image_pairs')


def stack_global_features(global_features_config: ImageFeatureConfig,
                          global_features_paths: List[Tuple[str, str]]
                          ) -> Tuple[Union[np.array, List[str]], np.ndarray]:
    """
    loads global features and store them inside a numpy array of shape (number_of_images, dsize)

    :param global_features_config: content of global_features.txt, required to load the global features
    :type global_features_config: ImageFeatureConfig
    :param global_features_paths: list of every image and the full path to the corresponding global feature
    :type global_features_paths: List[Tuple[str, str]]
    """
    logger.debug('loading global features')
    number_of_images = len(global_features_paths)

    stacked_features_index = np.empty((number_of_images,), dtype=object)
    stacked_features = np.empty((number_of_images, global_features_config.dsize), dtype=global_features_config.dtype)

    hide_progress_bar = logger.getEffectiveLevel() > logging.INFO
    for i, (image_path, global_feature_path) in tqdm(enumerate(global_features_paths), disable=hide_progress_bar):
        stacked_features_index[i] = image_path
        global_feature = image_global_features_from_file(global_feature_path,
                                                         global_features_config.dtype,
                                                         global_features_config.dsize)
        global_feature = global_feature / np.linalg.norm(global_feature)
        stacked_features[i] = global_feature

    return stacked_features_index, stacked_features


def compute_image_pairs(mapping_path: str,
                        query_path: str,
                        output_path: str,
                        topk: int):
    """
    compute image pairs between query -> mapping from global features, and write the result in a text file

    :param mapping_path: input path to kapture input root directory
    :type mapping_path: str
    :param query_path: input path to a kapture root directory
    :type query_path: str
    :param output_path: output path to pairsfile
    :type output_path: str
    :param topk: the max number of top retained images
    :type topk: int
    """
    logger.info(f'compute_image_pairs. loading mapping: {mapping_path}')
    kdata_mapping = kapture_from_dir(mapping_path)
    assert kdata_mapping.sensors is not None
    assert kdata_mapping.records_camera is not None
    assert kdata_mapping.global_features is not None

    if mapping_path == query_path:
        kdata_query = kdata_mapping
    else:
        logger.info(f'compute_image_pairs. loading query: {query_path}')
        kdata_query = kapture_from_dir(query_path)
        assert kdata_query.sensors is not None
        assert kdata_query.records_camera is not None
        assert kdata_query.global_features is not None

    assert kdata_mapping.global_features is not None
    assert kdata_query.global_features is not None
    assert kdata_mapping.global_features.type_name == kdata_query.global_features.type_name
    assert kdata_mapping.global_features.dtype == kdata_query.global_features.dtype
    assert kdata_mapping.global_features.dsize == kdata_query.global_features.dsize
    global_features_config = ImageFeatureConfig(kdata_mapping.global_features.type_name,
                                                kdata_mapping.global_features.dtype,
                                                kdata_mapping.global_features.dsize)

    logger.info(f'computing pairs from with {kdata_mapping.global_features.type_name}...')

    mapping_global_features_to_filepaths = global_features_to_filepaths(kdata_mapping.global_features,
                                                                        mapping_path)
    mapping_list = list(kapture.flatten(mapping_global_features_to_filepaths, is_sorted=True))
    mapping_indexes, mapping_features = stack_global_features(global_features_config,
                                                              mapping_list)

    if mapping_path == query_path:
        query_indexes, query_features = mapping_indexes, mapping_features
    else:
        query_global_features_to_filepaths = global_features_to_filepaths(kdata_query.global_features,
                                                                          query_path)
        query_list = list(kapture.flatten(query_global_features_to_filepaths, is_sorted=True))
        query_indexes, query_features = stack_global_features(global_features_config,
                                                              query_list)
    # compute similarity matrix
    similarity_matrix = query_features.dot(mapping_features.T)

    # convert similarity matrix to dictionary query_name -> sorted (high score first) list [(mapping_name, score), ...]
    similarity_dict = {}
    for i, line in enumerate(similarity_matrix):
        scores = line
        indexes = np.argsort(-scores)
        query_name = query_indexes[i]
        similarity_dict[query_name] = list(zip(mapping_indexes[indexes], scores[indexes]))

    # get list of image pairs
    image_pairs = []
    for query_image_name, images_to_match in sorted(similarity_dict.items()):
        for mapping_image_name, score in images_to_match[:topk]:
            image_pairs.append([query_image_name, mapping_image_name, score])

    logger.info('saving to file  ...')
    p = pathlib.Path(output_path)
    os.makedirs(str(p.parent.resolve()), exist_ok=True)
    with open(output_path, 'w') as fid:
        table_to_file(fid, image_pairs, header='# query_image, map_image, score')
    logger.info('all done')


def compute_image_pairs_command_line():
    parser = argparse.ArgumentParser(
        description=('Create image pairs files from global features. '
                     'Pairs are computed between query <-> mapping or mapping <-> mapping'))
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
                                  action=kapture.utils.logging.VerbosityParser,
                                  help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument('-q', '--silent', '--quiet',
                                  action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('--mapping', required=True,
                        help=('input path to kapture input root directory\n'
                              'it must contain global features for all images'))
    parser.add_argument('--query', required=True,
                        help=('input path to a kapture root directory containing query images\n'
                              'it must contain global features for all images\n'
                              'use the same value as mapping if you want to compute mapping <-> mapping matches'))
    parser.add_argument('-o', '--output', required=True,
                        help='output path to pairsfile')

    parser.add_argument('--topk',
                        default=20,
                        type=int,
                        help='the max number of top retained images')

    args = parser.parse_args()
    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    logger.debug(''.join(['\n\t{:13} = {}'.format(k, v)
                          for k, v in vars(args).items()]))
    compute_image_pairs(args.mapping, args.query, args.output, args.topk)


if __name__ == '__main__':
    compute_image_pairs_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Colmap command execution in sub processes
"""

import logging
import os
import os.path as path
import subprocess
from typing import List, Optional

import kapture
from kapture.utils.paths import safe_remove_file
from .database_extra import save_match_list

logger = logging.getLogger('colmap')


def run_colmap_command(colmap_binary_path: str, args: List[str]) -> None:
    """
    run any colmap command

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param args: list of arguments that will be passed to the colmap command
    :type args: List[str]
    :raises ValueError: colmap subprocess did not return 0
    """
    args.insert(0, colmap_binary_path)
    logger.info(args)
    colmap_process = subprocess.Popen(args)
    colmap_process.wait()

    if colmap_process.returncode != 0:
        raise ValueError(
            '\nSubprocess Error (Return code:'
            f' {colmap_process.returncode} )')


def run_feature_extractor(colmap_binary_path: str,
                          colmap_use_cpu: bool,
                          colmap_gpu_index: str,
                          colmap_db_path: str,
                          images_path: str,
                          image_list_path: str,
                          colmap_options: List[str] = None) -> None:
    """
    run colmap feature_extractor:
     Perform feature extraction or import features for a set of images

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param colmap_use_cpu: add --SiftExtraction.use_gpu 0
    :type colmap_use_cpu: bool
    :param colmap_gpu_index: add --SiftExtraction.gpu_index {colmap_gpu_index}
    :type colmap_gpu_index: str
    :param colmap_db_path: value for --database_path
    :type colmap_db_path: str
    :param images_path: value for --image_path
    :type images_path: str
    :param image_list_path: value for --image_list_path
    :type image_list_path: str
    :param colmap_options: list of additional parameters to add to the command, defaults to None
    :type colmap_options: List[str], optional
    """
    feature_args = ["feature_extractor",
                    "--database_path",
                    colmap_db_path,
                    "--image_path",
                    images_path,
                    "--image_list_path",
                    image_list_path]
    if colmap_options is not None and len(colmap_options) > 0:
        feature_args += colmap_options
    if colmap_use_cpu:
        feature_args += [
            "--SiftExtraction.use_gpu",
            "0"
        ]
    elif colmap_gpu_index:
        feature_args += [
            "--SiftExtraction.gpu_index",
            colmap_gpu_index
        ]

    run_colmap_command(colmap_binary_path, feature_args)


def run_vocab_tree_matcher(colmap_binary_path: str,
                           colmap_use_cpu: bool,
                           colmap_gpu_index: str,
                           colmap_db_path: str,
                           vocab_tree_path: str,
                           images_path: str = "") -> None:
    """
    run colmap vocab_tree_matcher:
     Perform feature matching after performing feature extraction

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param colmap_use_cpu: add --SiftExtraction.use_gpu 0
    :type colmap_use_cpu: bool
    :param colmap_gpu_index: add --SiftExtraction.gpu_index {colmap_gpu_index}
    :type colmap_gpu_index: str
    :param colmap_db_path: value for --database_path
    :type colmap_db_path: str
    :param vocab_tree_path: value for --VocabTreeMatching.vocab_tree_path
    :type vocab_tree_path: str
    :param images_path: value for --VocabTreeMatching.match_list_path, defaults to ""
    :type images_path: str, optional
    """
    vocab_tree_matcher_args = ["vocab_tree_matcher",
                               "--database_path",
                               colmap_db_path,
                               "--VocabTreeMatching.vocab_tree_path",
                               vocab_tree_path]
    if images_path != "":
        vocab_tree_matcher_args += ["--VocabTreeMatching.match_list_path", images_path]

    if colmap_use_cpu:
        vocab_tree_matcher_args += [
            "--SiftMatching.use_gpu",
            "0"
        ]
    elif colmap_gpu_index:
        vocab_tree_matcher_args += [
            "--SiftMatching.gpu_index",
            colmap_gpu_index
        ]
    run_colmap_command(colmap_binary_path, vocab_tree_matcher_args)


def run_matches_importer(colmap_binary_path: str,
                         colmap_use_cpu: bool,
                         colmap_gpu_index: Optional[str],
                         colmap_db_path: str,
                         match_list_path: str) -> None:
    """
    run colmap matches_importer:
     Perform geometric verification on matches

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param colmap_use_cpu: add --SiftExtraction.use_gpu 0
    :type colmap_use_cpu: bool
    :param colmap_gpu_index: add --SiftExtraction.gpu_index {colmap_gpu_index}
    :type colmap_gpu_index: str
    :param colmap_db_path: value for --database_path
    :type colmap_db_path: str
    :param match_list_path: value for --match_list_path
    :type match_list_path: str
    """
    matches_importer_args = ["matches_importer",
                             "--database_path",
                             colmap_db_path,
                             "--match_list_path",
                             match_list_path,
                             "--match_type",
                             'pairs']
    if colmap_use_cpu:
        matches_importer_args += [
            "--SiftMatching.use_gpu",
            "0"
        ]
    elif colmap_gpu_index:
        matches_importer_args += [
            "--SiftMatching.gpu_index",
            colmap_gpu_index
        ]
    run_colmap_command(colmap_binary_path, matches_importer_args)


def run_matches_importer_from_kapture(colmap_binary_path: str,
                                      colmap_use_cpu: bool,
                                      colmap_gpu_index: Optional[str],
                                      colmap_db_path: str,
                                      kapture_data: kapture.Kapture,
                                      force: bool = True,
                                      clean: bool = True) -> None:
    """
    export list of matches from kapture data then run colmap matches_importer

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param colmap_use_cpu: add --SiftExtraction.use_gpu 0
    :type colmap_use_cpu: bool
    :param colmap_gpu_index: add --SiftExtraction.gpu_index {colmap_gpu_index}
    :type colmap_gpu_index: str
    :param colmap_db_path: value for --database_path
    :type colmap_db_path: str
    :param kapture_data: kapture data that contains the matches (that are already in the colmap database) to verify
    :type kapture_data: kapture.Kapture
    :param force: do not ask before overwriting match_list.txt, defaults to True
    :type force: bool, optional
    :param clean: remove match_list.txt before exiting, defaults to True
    :type clean: bool, optional
    """
    db_dir = path.dirname(colmap_db_path)
    match_list_path = path.join(db_dir, 'match_list.txt')
    safe_remove_file(match_list_path, force)
    save_match_list(kapture_data, match_list_path)
    run_matches_importer(colmap_binary_path, colmap_use_cpu, colmap_gpu_index, colmap_db_path, match_list_path)
    if clean:
        os.remove(match_list_path)


def run_point_triangulator(colmap_binary_path: str,
                           colmap_db_path: str,
                           images_path: str,
                           input_path: str,
                           output_path: str,
                           point_triangulator_options: List[str]) -> None:
    """
    run colmap point_triangulator:
     Triangulate all observations of registered images in an existing model using the feature matches in a database

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param colmap_db_path: value for --database_path
    :type colmap_db_path: str
    :param images_path: value for --image_path
    :type images_path: str
    :param input_path: value for --input_path
    :type input_path: str
    :param output_path: value for --output_path
    :type output_path: str
    :param point_triangulator_options: list of additional parameters to add to the command
    :type point_triangulator_options: List[str]
    """
    point_triangulator_args = ["point_triangulator",
                               "--database_path",
                               colmap_db_path,
                               "--image_path",
                               images_path,
                               "--input_path",
                               input_path,
                               "--output_path",
                               output_path]
    if point_triangulator_options is not None and len(point_triangulator_options) > 0:
        point_triangulator_args += point_triangulator_options
    run_colmap_command(colmap_binary_path, point_triangulator_args)


def run_mapper(colmap_binary_path: str,
               colmap_db_path: str,
               images_path: str,
               input_path: Optional[str],
               output_path: str,
               mapper_options: List[str]) -> None:
    """
    run colmap mapper:
     Sparse 3D reconstruction / mapping of the dataset using SfM after performing feature extraction and matching

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param colmap_db_path: value for --database_path
    :type colmap_db_path: str
    :param images_path: value for --image_path
    :type images_path: str
    :param input_path: value for --input_path
    :type input_path: Optional[str]
    :param output_path: value for --output_path
    :type output_path: str
    :param mapper_options: list of additional parameters to add to the command
    :type mapper_options: List[str]
    """
    mapper_args = ["mapper",
                   "--database_path",
                   colmap_db_path,
                   "--image_path",
                   images_path,
                   "--output_path",
                   output_path]
    if input_path is not None:
        mapper_args += [
            "--input_path",
            input_path]
    if mapper_options is not None and len(mapper_options) > 0:
        mapper_args += mapper_options
    run_colmap_command(colmap_binary_path, mapper_args)


def run_bundle_adjustment(colmap_binary_path: str,
                          input_path: str,
                          output_path: str,
                          output_rig_path: str = "") -> None:
    """
    run colmap bundle_adjuster or colmap rig_bundle_adjuster (if output_rig_path is provided)

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param input_path: value for --input_path
    :type input_path: str
    :param output_path: value for --output_path
    :type output_path: str
    :param output_rig_path: value for --rig_config_path, if set, run rig_bundle_adjuster instead of bundle_adjuster
    :type output_rig_path: str, optional
    """
    if output_rig_path:
        logging.info("Run bundle adjuster with rig")
        args = ["rig_bundle_adjuster", "--rig_config_path", output_rig_path]
    else:
        args = ["bundle_adjuster"]
    args.extend(["--input_path", input_path,
                 "--output_path", output_path])
    run_colmap_command(colmap_binary_path, args)


def run_image_registrator(colmap_binary_path: str,
                          colmap_db_path: str,
                          input_path: str,
                          output_path: str,
                          image_registrator_options: List[str]) -> None:
    """
    run colmap image_registrator:
     Register new images in the database against an existing model

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param colmap_db_path: value for --database_path
    :type colmap_db_path: str
    :param input_path: value for --input_path
    :type input_path: str
    :param output_path: value for --output_path
    :type output_path: str
    :param image_registrator_options: list of additional parameters to add to the command
    :type image_registrator_options: List[str]
    """
    image_registrator_args = ["image_registrator",
                              "--database_path",
                              colmap_db_path,
                              "--input_path",
                              input_path,
                              "--output_path",
                              output_path]
    if image_registrator_options is not None and len(image_registrator_options) > 0:
        image_registrator_args += image_registrator_options
    run_colmap_command(colmap_binary_path, image_registrator_args)


def run_model_converter(colmap_binary_path: str,
                        input_path: str,
                        output_path: str) -> None:
    """
    run colmap model_converter with --output_type TXT:
     convert reconstruction from binary files to TXT files

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param input_path: value for --input_path
    :type input_path: str
    :param output_path: value for --output_path
    :type output_path: str
    """
    model_converter_args = ["model_converter",
                            "--input_path",
                            input_path,
                            "--output_path",
                            output_path,
                            "--output_type",
                            "TXT"]
    run_colmap_command(colmap_binary_path, model_converter_args)


def run_image_undistorter(colmap_binary_path: str,
                          image_path: str,
                          input_path: str,
                          output_path: str) -> None:
    """
    run colmap image_undistorter:
     Undistort images and/or export them for MVS or to external dense reconstruction software, such as CMVS/PMVS

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param image_path: value for -image_path
    :type image_path: str
    :param input_path: value for --input_path
    :type input_path: str
    :param output_path: value for --output_path
    :type output_path: str
    """
    image_undistorter_args = ["image_undistorter",
                              "--image_path",
                              image_path,
                              "--input_path",
                              input_path,
                              "--output_path",
                              output_path]
    run_colmap_command(colmap_binary_path, image_undistorter_args)


def run_patch_match_stereo(colmap_binary_path: str,
                           workspace_path: str) -> None:
    """
    run colmap patch_match_stereo:
     Dense 3D reconstruction / mapping using MVS after running the image_undistorter to initialize the workspace

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param workspace_path: value for --workspace_path
    :type workspace_path: str
    """
    patch_match_stereo_args = ["patch_match_stereo",
                               "--workspace_path",
                               workspace_path]
    run_colmap_command(colmap_binary_path, patch_match_stereo_args)


def run_stereo_fusion(colmap_binary_path: str,
                      workspace_path: str,
                      output_path: str) -> None:
    """
    run colmap stereo_fusion

    :param colmap_binary_path: path to colmap executable
    :type colmap_binary_path: str
    :param workspace_path: value for --workspace_path
    :type workspace_path: str
    :param output_path: value for --output_path
    :type output_path: str
    """
    stereo_fusion_args = ["stereo_fusion",
                          "--workspace_path",
                          workspace_path,
                          "--output_path",
                          output_path]
    run_colmap_command(colmap_binary_path, stereo_fusion_args)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
script to easily  download kapture datasets.
"""

import argparse
import logging
import sys
import os
import os.path as path
import requests
import yaml
import fnmatch
from typing import Dict, Optional, List
from tqdm import tqdm
import tarfile
import hashlib
from shutil import rmtree
from datetime import datetime

logger = logging.getLogger('download')
logging.basicConfig(format='%(levelname)-8s::%(name)s: %(message)s')

INDEX_FILENAME = 'kapture_dataset_index.yaml'
DEFAULT_DATASET_PATH = path.normpath(path.abspath('.'))
DEFAULT_REPOSITORY_URL = 'https://github.com/naver/kapture/raw/master/dataset'

datasets = {}


def ask_confirmation(question):
    """
    ask "question" to the user.
    The "answer" return value is True for "yes" or False for "no".
    """
    validate = ['yes', 'y', 'ye']
    prompt = ' [y/N]\n'
    sys.stdout.write(question + prompt)
    user_choice = input().lower()
    return user_choice in validate


class Dataset:
    def __init__(self,
                 name: str,
                 url: str,
                 root_path: str,
                 sha256sum: str,
                 sub_path: Optional[str] = None
                 ):
        """
        :param name: name of the archive (dataset or part of a dataset)
        :param url: input remote url of the dataset archive (tar).
        :param root_path: input absolute path to root directory where all datasets are installed.
        :param sha256sum: sha256 sum of the archive file on remote server.
        :param sub_path: input optional sub path where to install the dataset under root_path/sub_path.
                         If not given, tar file is extracted directly to root_path
        """
        self._name = name
        # remore url of the archive
        self._archive_url = url
        # local absolute path to archive
        self._archive_filepath = path.join(root_path, name + '.tar')
        # if _success_filepath is present, it means it as already been installed. It contains some debug info
        self._success_filepath = path.join(root_path, name + '.txt')
        # local absolute path where to the install the dataset (multiple dataset might be installed at same path).
        self._install_local_path = path.join(root_path, sub_path) if sub_path is not None else root_path
        self._sha256sum_archive_remote = sha256sum

    def is_archive_valid(self):
        if not path.isfile(self._archive_filepath):
            return False
        # size is consistent, check sha256
        sha256_hash = hashlib.sha256()
        with open(self._archive_filepath, 'rb') as f:
            # Read and update hash string value in blocks of 4K
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
        sha256sum_archive_local = sha256_hash.hexdigest()
        logger.debug(f'sha256sum {self._archive_filepath} :\n\t{sha256sum_archive_local}')
        if sha256sum_archive_local != self._sha256sum_archive_remote:
            return False

        return True

    def prob_status(self):
        """
        gives the actual dataset status
         - online: means not downloaded, and not installed (extracted).
         - installed: means has been downloaded and installed (extracted).
         - downloaded: means has been downloaded (tar) but not installed (extracted) yet.
         - incomplete: means partially downloaded
         - corrupted: means that the downloaded archive is curropted (inconsistent size or sh256).
        """
        if path.isfile(self._success_filepath):
            return 'installed'
        if not path.isfile(self._archive_filepath):
            return 'online'

        assert path.isfile(self._archive_filepath)
        # file exists, check it has proper size
        r = requests.head(self._archive_url)
        size_archive_online = int(r.headers.get('content-length', 0))
        size_archive_local = int(path.getsize(self._archive_filepath))
        if size_archive_local < size_archive_online:
            logger.debug(f'file_size_online={size_archive_online} != file_size_local={size_archive_local}')
            return 'incomplete'
        if size_archive_local > size_archive_online:
            logger.critical(f'inconsistent file size ({size_archive_online} vs {size_archive_local})')
            return 'corrupted'

        assert size_archive_local == size_archive_online
        # size is consistent, check sha256
        if not self.is_archive_valid():
            return 'corrupted'

        # archive is there, but not installed
        return 'downloaded'

    def __repr__(self):
        return f'{self._name:30} | {self.prob_status():10} | {self._archive_url}'

    def download_archive_resume(self, resume_byte_pos: Optional[int] = None):
        """
        resume (or start if no pos given) the dataset download.
        :param resume_byte_pos: input position in bytes where to resume the Download
        """
        r = requests.head(self._archive_url)
        file_size_online = int(r.headers.get('content-length', 0))
        # Append information to resume download at specific byte position to header
        resume_header = ({'Range': f'bytes={resume_byte_pos}-'}
                         if resume_byte_pos else None)
        # Establish connection
        r = requests.get(self._archive_url, stream=True, headers=resume_header)
        # Set configuration
        block_size = 1024
        initial_pos = resume_byte_pos if resume_byte_pos else 0
        mode = 'ab' if resume_byte_pos else 'wb'
        with open(self._archive_filepath, mode) as f:
            with tqdm(total=file_size_online, unit='B',
                      unit_scale=True, unit_divisor=block_size,
                      desc=self._archive_filepath, initial=initial_pos,
                      ascii=True, miniters=1,
                      disable=logger.getEffectiveLevel() >= logging.CRITICAL) as pbar:
                for chunk in r.iter_content(32 * block_size):
                    f.write(chunk)
                    pbar.update(len(chunk))

    def download_archive_file(self):
        """ Starts or resumes the download if already started. """
        logger.debug(f'downloading {self._archive_filepath}')
        r = requests.head(self._archive_url)
        if path.isfile(self._archive_filepath):
            logger.debug('file is already (partially) there.')
            # file already there (at least partially)
            file_size_online = int(r.headers.get('content-length', 0))
            file_size_local = int(path.getsize(self._archive_filepath))
            if file_size_online == file_size_local:
                logger.info(f'file {self._archive_filepath} already downloaded.')
            else:
                logger.info(f'resume download from {file_size_local / file_size_online * 100.:4.1f}%')
                self.download_archive_resume(file_size_local)
        else:
            logger.info(f'start downloading {self._archive_filepath}.')
            self.download_archive_resume()

    def untar_archive_file(self):
        if not self.is_archive_valid():
            raise ValueError('Archive file not valid: cant install.')
        logger.debug(f'extracting\n\tfrom: {self._archive_filepath}\n\tto  : {self._install_local_path}')
        # make sure directory exists
        os.makedirs(self._install_local_path, exist_ok=True)
        with tarfile.open(self._archive_filepath, 'r:*') as archive:
            archive.extractall(self._install_local_path)
        # creating success file
        logger.debug(f'creating {self._success_filepath}')
        with open(self._success_filepath, 'wt') as f:
            f.write(f'installed on: {datetime.now()}\nfrom: {self._archive_url}')
        # cleaning tar
        logger.debug(f'cleaning {self._archive_filepath}')
        os.remove(self._archive_filepath)

    def install(self, force_overwrite: bool = False):
        """ Install handle download and untar """
        # test the dataset presence
        current_status = self.prob_status()
        if current_status == 'installed' and not force_overwrite:
            logger.info(f'{self._install_local_path} already exists: skipped')
            return

        if current_status != 'downloaded':
            # check archive file integrity
            if current_status == 'corrupted':
                # if corrupted: remove the archive and start over
                os.remove(self._archive_filepath)
            self.download_archive_file()

        # install it
        logger.info(f'installing {path.basename(self._archive_filepath)} to {self._install_local_path}')
        self.untar_archive_file()
        # done
        logger.info(f'done isntalling {self._name}')

    def clean(self):
        """
        Remove all files (installed, archive, ..)
        warning: this may clean also other intricate installation

        :return:
        """
        # clean archive if any
        if path.isfile(self._archive_filepath):
            logger.debug(f'deleting {self._archive_filepath}')
            os.remove(self._archive_filepath)
        # clean success files if any
        if path.isfile(self._success_filepath):
            logger.debug(f'deleting {self._success_filepath}')
            os.remove(self._success_filepath)
        # clean installation if any
        if path.isdir(self._install_local_path):
            # warning: this may clean also other intricate installation
            logger.debug(f'deleting {self._install_local_path}/')
            # TODO: check symlink attack: https://bugs.python.org/issue4489
            rmtree(self._install_local_path)


def load_index(
        index_filepath: str,
        root_path: str,
        filter_patterns: Optional[List[str]] = None
) -> Dict[str, Dataset]:
    """
    Parses and load data from the index files, under yaml format.
    the yaml file looks like :
    ----
    robotcar_seasons_02:
      sub_path: robotcar_seasons
      url: http://download.europe.naverlabs.com//kapture/robotcar_seasons_02.tar
      sha256sum: 542ef47c00d5e387cfb0dcadb2459ae2fb17d59010cc51bae0c49403b4fa6a18
    ----

    :param index_filepath: input absolute path to index file
    :param root_path: input absolute path to install directory
    :param filter_patterns: optional input list of unix-like patterns (e.g. SiLDa*) to filter datasets
    :return: dict name -> [url, sub_path, sha256sum]
    """
    if not path.isfile(index_filepath):
        raise FileNotFoundError('no index file: do an update.')
    with open(index_filepath, 'rt') as f:
        dataset_yaml = yaml.safe_load(f)
    if len(dataset_yaml) == 0:
        raise FileNotFoundError('invalid index file: do an update.')

    datasets = {
        name: Dataset(name=name, root_path=root_path, **data) for name, data in dataset_yaml.items()
        if filter_patterns is None or any(fnmatch.fnmatch(name, pattern) for pattern in filter_patterns)
    }
    return datasets


def kapture_dataset_download_cli():
    """
    Parse the kapture_dataset_download command line .
    """
    parser = argparse.ArgumentParser(description='download kapture datasets.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', action='store_const',
                                  dest='verbose', const=logging.INFO,
                                  help='display info messages [True].')
    parser_verbosity.add_argument('-q', '--silent', '--quiet', action='store_const',
                                  dest='verbose', const=logging.CRITICAL,
                                  help='silence all messages.')
    parser_verbosity.add_argument('-d', '--debug', action='store_const',
                                  dest='verbose', const=logging.DEBUG,
                                  help='display info and debug messages.')
    parser.add_argument('--install_path', default=path.normpath(path.join(DEFAULT_DATASET_PATH)),
                        help=f'path to index files listing all datasets'
                             f' [{path.normpath(path.join(DEFAULT_DATASET_PATH))}]')
    subparsers = parser.add_subparsers(help='sub-command help')
    ####################################################################################################################
    # create the parser for the "update" command
    parser_update = subparsers.add_parser('update', help='update dataset index')
    parser_update.set_defaults(cmd='update')
    parser_update.add_argument('--repo', default=DEFAULT_REPOSITORY_URL,
                               help='url of the repository.')
    ####################################################################################################################
    parser_list = subparsers.add_parser('list', help='display dataset index')
    parser_list.set_defaults(cmd='list')
    parser_list.add_argument('dataset', nargs='*', default=['*'])
    ####################################################################################################################
    parser_install = subparsers.add_parser('install', help='install dataset')
    parser_install.set_defaults(cmd='install')
    parser_install.add_argument('-f', '--force', action='store_true', default=False,
                                help='Force installation even if dataset has already been installed.')
    parser_install.add_argument('dataset', nargs='*', default=['+'],
                                help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################
    parser_download = subparsers.add_parser('download', help='dowload dataset, without installing it')
    parser_download.set_defaults(cmd='download')
    parser_download.add_argument('dataset', nargs='*', default=['+'],
                                 help='name of the dataset to download. Can use unix-like wildcard.')
    ####################################################################################################################
    parser_clean = subparsers.add_parser('clean', help='clean all')
    parser_clean.set_defaults(cmd='clean')
    parser_clean.add_argument('-f', '--force', action='store_true', default=False,
                              help='Do not ask for user confirmation.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose or logging.INFO)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    try:
        index_filepath = path.join(args.install_path, INDEX_FILENAME)
        if not hasattr(args, 'cmd'):
            # check user did not forgot the command
            logger.critical(f'Choose command among [ {" | ".join(subparsers.choices)} ]')
            exit(-1)

        if args.cmd == 'update':
            logger.info(f'updating dataset list from {args.repo} ...')
            index_remote_url = path.join(args.repo, INDEX_FILENAME)
            logger.debug(f'retrieving index at {index_remote_url}')
            r = requests.get(index_remote_url, allow_redirects=True)
            if r.status_code != requests.codes.ok:
                raise ConnectionError(f'unable to grab {index_remote_url} (code:{r.status_code})')
            with open(index_filepath, 'wt') as f:
                f.write(r.text)
            logger.info('dataset index retrieved successfully.')

        elif args.cmd == 'list':
            logger.info(f'listing dataset {index_filepath} ...')
            dataset_index = load_index(index_filepath=index_filepath,
                                       root_path=args.install_path,
                                       filter_patterns=args.dataset)
            print('\n'.join(str(dataset) for dataset in dataset_index.values()))

        elif args.cmd == 'install':
            logger.info(f'installing dataset {args.dataset} ...')
            dataset_index = load_index(index_filepath=index_filepath,
                                       root_path=args.install_path,
                                       filter_patterns=args.dataset)
            if len(dataset_index) == 0:
                raise ValueError('no matching dataset')
            logger.info(f'{len(dataset_index)} dataset will be installed.')
            for name, dataset in dataset_index.items():
                logger.info(f'downloading {name} ...')
                dataset.install(force_overwrite=args.force)

        elif args.cmd == 'download':
            logger.info(f'downloading dataset {args.dataset} ...')
            dataset_index = load_index(index_filepath=index_filepath,
                                       root_path=args.install_path,
                                       filter_patterns=args.dataset)
            if len(dataset_index) == 0:
                raise ValueError('no matching dataset')
            logger.info(f'{len(dataset_index)} dataset will be downloaded.')
            for name, dataset in dataset_index.items():
                logger.info(f'downloading {name} ...')
                dataset.download_archive_file()

        elif args.cmd == 'clean':
            logger.info(f'cleaning all dataset ...')
            # since we'll remove everything, ask user confirmation
            dataset_index = load_index(index_filepath=index_filepath,
                                       root_path=args.install_path)
            # just for info, enumerate installed or partially installed = not online
            nb_dataset_isntalled_index = len([d for d in dataset_index.values()
                                              if d.prob_status() is not 'online'])
            logger.info(f'all ({nb_dataset_isntalled_index}) datasets will be erased .')
            if not args.force and not ask_confirmation(
                    f'Are you sure you want to delete ALL {len(dataset_index)} datasets '):
                logger.info('cleaning canceled by user')
            else:
                for name, dataset in dataset_index.items():
                    logger.info(f'cleaning {name} ...')
                    dataset.clean()

        else:
            raise ValueError(f'unknown command {args.cmd}')

    except Exception as e:
        logger.critical(e)


if __name__ == '__main__':
    kapture_dataset_download_cli()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Openmvg constants values useful for conversion from and to.
"""

from kapture.utils.Collections import AutoEnum
from enum import auto

DEFAULT_JSON_FILE_NAME = 'sfm_data.json'
SFM_DATA_VERSION_NUMBER = "0.3"

# XML names
SFM_DATA_VERSION = "sfm_data_version"
ROOT_PATH = "root_path"
VIEWS = "views"
VIEW_PRIORS = "view_priors"
KEY = "key"
VALUE = "value"
POLYMORPHIC_ID = "polymorphic_id"
PTR_WRAPPER = "ptr_wrapper"
ID = "id"
DATA = "data"
LOCAL_PATH = "local_path"
FILENAME = "filename"
WIDTH = "width"
HEIGHT = "height"
ID_VIEW = "id_view"
ID_INTRINSIC = "id_intrinsic"
ID_POSE = "id_pose"
INTRINSICS = "intrinsics"
POLYMORPHIC_NAME = "polymorphic_name"
VALUE0 = "value0"
FOCAL_LENGTH = "focal_length"
PRINCIPAL_POINT = "principal_point"
DISTO_K1 = "disto_k1"
DISTO_K3 = "disto_k3"
DISTO_T2 = "disto_t2"
FISHEYE = "fisheye"
EXTRINSICS = "extrinsics"
USE_POSE_ROTATION_PRIOR = "use_pose_rotation_prior"
ROTATION_WEIGHT = "rotation_weight"
ROTATION = "rotation"
USE_POSE_CENTER_PRIOR = "use_pose_center_prior"
CENTER_WEIGHT = "center_weight"
CENTER = "center"
STRUCTURE = "structure"
CONTROL_POINTS = "control_points"


# Camera models
class CameraModel(AutoEnum):
    """
    Enumeration that contains all the openmvg camera model
    """
    pinhole = auto()
    pinhole_radial_k1 = auto()
    pinhole_radial_k3 = auto()
    pinhole_brown_t2 = auto()
    fisheye = auto()



import os
import logging
import os.path as path
import numpy as np
import quaternion
import gzip
import pickle
import json
from tqdm import tqdm
from typing import Optional, Dict
# kapture
import kapture
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.io.features import get_keypoints_fullpath, image_keypoints_from_file
from kapture.io.features import get_descriptors_fullpath, image_descriptors_from_file
from kapture.io.features import get_matches_fullpath, image_matches_from_file
from kapture.io.records import TransferAction, transfer_files_from_dir, get_record_fullpath

logger = logging.getLogger('opensfm')


"""
opensfm_project/
├── config.yaml
├── images/
├── masks/
├── gcp_list.txt
├── exif/
├── camera_models.json
├── features/
├── matches/
├── tracks.csv
├── reconstruction.json
├── reconstruction.meshed.json
└── undistorted/
    ├── images/
    ├── masks/
    ├── tracks.csv
    ├── reconstruction.json
    └── depthmaps/
        └── merged.ply
"""

"""
reconstruction.json: [RECONSTRUCTION, ...]

RECONSTRUCTION: {
    "cameras": {
        CAMERA_ID: CAMERA,
        ...
    },
    "shots": {
        SHOT_ID: SHOT,
        ...
    },
    "points": {
        POINT_ID: POINT,
        ...
    }
}

CAMERA: {
    "projection_type": "perspective",  # Can be perspective, brown, fisheye or equirectangular
    "width": NUMBER,                   # Image width in pixels
    "height": NUMBER,                  # Image height in pixels

    # Depending on the projection type more parameters are stored.
    # These are the parameters of the perspective camera.
    "focal": NUMBER,                   # Estimated focal length
    "k1": NUMBER,                      # Estimated distortion coefficient
    "k2": NUMBER,                      # Estimated distortion coefficient
}

SHOT: {
    "camera": CAMERA_ID,
    "rotation": [X, Y, Z],      # Estimated rotation as an angle-axis vector
    "translation": [X, Y, Z],   # Estimated translation
    "gps_position": [X, Y, Z],  # GPS coordinates in the reconstruction reference frame
    "gps_dop": METERS,          # GPS accuracy in meters
    "orientation": NUMBER,      # EXIF orientation tag (can be 1, 3, 6 or 8)
    "capture_time": SECONDS     # Capture time as a UNIX timestamp
}

POINT: {
    "coordinates": [X, Y, Z],      # Estimated position of the point
    "color": [R, G, B],            # Color of the point
}
"""

"""
reconstruction.meshed.json
[{
    'cameras': {'v2 unknown unknown 1920 1080 perspective 0': 
                {'projection_type': 'perspective', 'width': 1920, 'height': 1080, 'focal': 0.8647151305270488, 
                'k1': 0.04060214391621549, 'k2': -0.04060273810852096}},
    'shots': {
        'frame00016.png' : {
        'rotation': [1.5061234719524716, 0.06688721174244067, -0.030847050348337724],
         'translation': [-2.1535823328020456, 0.28345212194377944, 1.2491740134158436], 
         'camera': 'v2 unknown unknown 1920 1080 perspective 0', 
         'orientation': 1,
         'capture_time': 0.0,
         'gps_dop', 'gps_position',
         'vertices',     # only in meshed
         'faces' [[x, y, z, ...]  # only in meshed
         },
         ...
    },
     'points': {
        '1': {
            'color': [74.0, 43.0, 31.0],
            'coordinates': [-4.204454761953588, 11.796709404713068, 4.7276044200915]
        },
        ...
     },
     'reference_lla': {
        'latitude': 0.0,
        'longitude': 0.0,
        'altitude': 0
     }
}]
"""


def export_opensfm_camera(
        kapture_camera: kapture.Camera,
        name: Optional[str] = None
) -> Dict[str, float]:
    """
    Converts kapture camera to OpenSfM.
    OpenSfm propose 3 models of camera: perspective, fisheye, spherical.
    Perspective Camera of OpenSfM
    (see https://www.opensfm.org/docs/geometry.html#camera-models)

    OpenSfM camera axis:
        - The z-axis points forward
        - The y-axis points down
        - The x-axis points to the right

    OpenSfM Pose:
        OpenSfM, however, chooses not to store the “camera origin” in Pose objects.
        Instead, it stores the camera coordinates of the world origin in the translation field.
        Meaning Camera from World transformation.
        - rotation is represented as axis-angle vector.

    :param kapture_camera:
    :param name:
    :return:
    """
    corresponding_opensfm_perspective = [
        kapture.CameraType.SIMPLE_PINHOLE,
        kapture.CameraType.RADIAL,
        kapture.CameraType.SIMPLE_RADIAL
    ]

    if kapture_camera.camera_type not in corresponding_opensfm_perspective:
        raise ValueError(f'Unable to export camera of type "{kapture_camera.camera_type}" to OpenSfM')

    # SIMPLE_PINHOLE = w, h, fx, fy, cx, cy
    # SIMPLE_RADIAL  = w, h, f, cx, cy, k1
    # RADIAL         = w, h, f, cx, cy, k1, k2

    # note that principal point is ignred by openSfM
    largest_side_in_pixels = max(kapture_camera.camera_params[0], kapture_camera.camera_params[1])
    opensfm_camera = {
        'projection_type': 'perspective',
        'width': int(kapture_camera.camera_params[0]),
        'height': int(kapture_camera.camera_params[1]),
        # The focal length provided by the EXIF metadata divided by the sensor width
        'focal': kapture_camera.camera_params[2] / largest_side_in_pixels,
        'k1': 0.0,
        'k2': 0.0,
    }

    # Distortion model of OpenSfM match CameraType.RADIAL
    if kapture_camera.camera_type in [kapture.CameraType.SIMPLE_RADIAL, kapture.CameraType.RADIAL]:
        opensfm_camera['k1'] = kapture_camera.camera_params[5]

    if kapture_camera.camera_type == kapture.CameraType.RADIAL:
        opensfm_camera['k2'] = kapture_camera.camera_params[6]

    return opensfm_camera


def export_opensfm(
        kapture_rootdir: str,
        opensfm_rootdir: str,
        force_overwrite_existing: bool = False,
        images_import_method: TransferAction = TransferAction.copy
) -> None:
    """

    :param kapture_rootdir:
    :param opensfm_rootdir:
    :param force_overwrite_existing:
    :param images_import_method:
    :return:
    """

    disable_tqdm = logger.getEffectiveLevel() > logging.INFO  # dont display tqdm for non-verbose levels
    # load reconstruction
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_dirpath=kapture_rootdir)

    # export cameras
    opensfm_cameras = {}
    kapture_cameras = {cam_id: cam
                       for cam_id, cam in kapture_data.sensors.items()
                       if cam.sensor_type == 'camera'}
    for cam_id, kapture_camera in kapture_cameras.items():
        opensfm_cameras[cam_id] = export_opensfm_camera(kapture_camera)

    # export shots
    opensfm_shots = {}
    for timestamp, camera_id, image_filename in tqdm(kapture.flatten(kapture_data.records_camera), disable=disable_tqdm):
        # retrieve pose (if there is one).
        # opensfm_shots = {image_filename: shot}
        # shot = {camera , rotation, translation, capture_time, gps_position, ...}
        opensfm_shot = {
            'capture_time': 0,  # in ms != timestamp
            'camera': camera_id,
        }
        if (timestamp, camera_id) in kapture_data.trajectories:
            pose = kapture_data.trajectories[timestamp, camera_id]
            rotation_vector = quaternion.as_rotation_vector(pose.r)
            translation_vector = pose.t.flatten()
            opensfm_shot.update({
                'rotation': rotation_vector.tolist(),
                'translation': translation_vector.tolist()
            })
        opensfm_shots[image_filename] = opensfm_shot

    # pack it
    opensfm_reconstruction = {
        'cameras': opensfm_cameras,
        'shots': opensfm_shots,
    }

    # images
    logger.info(f'writing image files "{path.join(opensfm_rootdir, "images")}".')
    image_filenames = [f for _, _, f in kapture.flatten(kapture_data.records_camera)]
    kapture_image_filepaths = [get_record_fullpath(kapture_rootdir, image_filename)
                               for image_filename in image_filenames]
    opensfm_image_filepaths = [path.join(opensfm_rootdir, 'images', image_filename)
                               for image_filename in image_filenames]
    transfer_files_from_dir(
        source_filepath_list=kapture_image_filepaths,
        destination_filepath_list=opensfm_image_filepaths,
        force_overwrite=force_overwrite_existing,
        copy_strategy=images_import_method,
    )

    # export features files (keypoints + descriptors)
    opensfm_features_suffix = '.features.npz'
    opensfm_features_dirpath = path.join(opensfm_rootdir, 'features')
    logger.info(f'exporting keypoint and descriptors to {opensfm_features_dirpath}')
    os.makedirs(opensfm_features_dirpath, exist_ok=True)
    for image_filename in tqdm(image_filenames, disable=disable_tqdm):
        opensfm_features = {}
        # look and load for keypoints in kapture
        if kapture_data.keypoints is not None and image_filename in kapture_data.keypoints:
            kapture_keypoints_filepath = get_keypoints_fullpath(kapture_dirpath=kapture_rootdir,
                                                                image_filename=image_filename)
            logger.debug(f'loading {kapture_keypoints_filepath}')
            kapture_keypoint = image_keypoints_from_file(kapture_keypoints_filepath,
                                                         dtype=kapture_data.keypoints.dtype,
                                                         dsize=kapture_data.keypoints.dsize)
            opensfm_features['points'] = kapture_keypoint

        # look and load for descriptors in kapture
        if kapture_data.descriptors is not None and image_filename in kapture_data.descriptors:
            kapture_descriptor_filepath = get_descriptors_fullpath(kapture_dirpath=kapture_rootdir,
                                                                   image_filename=image_filename)
            logger.debug(f'loading {kapture_descriptor_filepath}')
            kapture_descriptor = image_descriptors_from_file(kapture_descriptor_filepath,
                                                             dtype=kapture_data.descriptors.dtype,
                                                             dsize=kapture_data.descriptors.dsize)
            opensfm_features['descriptors'] = kapture_descriptor

        # writing opensfm feature file
        if len(opensfm_features) > 0:
            opensfm_features_filepath = path.join(opensfm_features_dirpath, image_filename + opensfm_features_suffix)
            logger.debug(f'writing {opensfm_features_filepath}')
            os.makedirs(path.dirname(opensfm_features_filepath), exist_ok=True)
            np.save(opensfm_features_filepath, opensfm_features)

    # export matches files
    if kapture_data.matches is not None:
        opensfm_matches_suffix = '_matches.pkl.gz'
        opensfm_matches_dirpath = path.join(opensfm_rootdir, 'matches')
        os.makedirs(opensfm_matches_dirpath, exist_ok=True)
        logger.info(f'exporting matches to {opensfm_matches_dirpath}')
        opensfm_pairs = {}
        for image_filename1, image_filename2 in kapture_data.matches:
            opensfm_pairs.setdefault(image_filename1, []).append(image_filename2)

        for image_filename1 in tqdm(image_filenames, disable=disable_tqdm):
            opensfm_matches = {}
            opensfm_matches_filepath = path.join(opensfm_matches_dirpath, image_filename1 + opensfm_matches_suffix)
            logger.debug(f'loading matches for {image_filename1}')
            for image_filename2 in opensfm_pairs.get(image_filename1, []):
                # print(image_filename1, image_filename2)
                kapture_matches_filepath = get_matches_fullpath(
                    (image_filename1, image_filename2), kapture_dirpath=kapture_rootdir)
                kapture_matches = image_matches_from_file(kapture_matches_filepath)
                opensfm_matches[image_filename2] = kapture_matches[:, 0:2].astype(np.int)

            os.makedirs(path.dirname(opensfm_matches_filepath), exist_ok=True)
            with gzip.open(opensfm_matches_filepath, 'wb') as f:
                pickle.dump(opensfm_matches, f)

    # export 3D-points files
    if kapture_data.points3d is not None:
        logger.info('exporting points 3-D')
        opensfm_reconstruction['points'] = {}
        for i, (x, y, z, r, g, b) in tqdm(enumerate(kapture_data.points3d), disable=disable_tqdm):
            opensfm_reconstruction['points'][i] = {
                'coordinates': [x, y, z],
                'color': [r, g, b]
            }

    # write json files #################################################################################################
    os.makedirs(opensfm_rootdir, exist_ok=True)
    # write reconstruction.json
    opensfm_reconstruction_filepath = path.join(opensfm_rootdir, 'reconstruction.json')
    logger.info(f'writing reconstruction file "{opensfm_reconstruction_filepath}".')
    with open(opensfm_reconstruction_filepath, 'wt') as f:
        json.dump([opensfm_reconstruction], f, indent=4)

    # write camera_models.json
    opensfm_cameras_filepath = path.join(opensfm_rootdir, 'camera_models.json')
    logger.info(f'writing camera models file "{opensfm_cameras_filepath}".')
    with open(opensfm_cameras_filepath, 'wt') as f:
        json.dump(opensfm_cameras, f, indent=4)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import argparse
import logging
from functools import lru_cache
from typing import List, Tuple
import torch
from tqdm import tqdm
from collections import OrderedDict

import path_to_kapture
import kapture
import kapture.utils.logging
from kapture.io.csv import kapture_from_dir, table_from_file
from kapture.algo.matching import MatchPairNnTorch
from kapture.io.features import get_descriptors_fullpath, get_matches_fullpath
from kapture.io.features import image_descriptors_from_file
from kapture.io.features import matches_check_dir, image_matches_to_file

logger = logging.getLogger('compute_matches')


def get_pairs_from_file(pairsfile_path: str) -> List[Tuple[str, str]]:
    """
    read a pairs file (csv with 3 fields, name1, name2, score) and return the list of matches

    :param pairsfile_path: path to pairsfile
    :type pairsfile_path: str
    """
    logger.info('reading pairs from pairsfile')
    image_pairs = []
    with open(pairsfile_path, 'r') as fid:
        table = table_from_file(fid)
        for query_name, map_name, _ in table:  # last field score is not used
            if query_name != map_name:
                image_pairs.append((query_name, map_name) if query_name < map_name else (map_name, query_name))
    # remove duplicates without breaking order
    image_pairs = list(OrderedDict.fromkeys(image_pairs))
    return image_pairs


@lru_cache(maxsize=50)
def load_descriptors(input_path: str, image_name: str, dtype, dsize):
    """
    load a descriptor. this functions caches up to 50 descriptors

    :param input_path: input path to kapture input root directory
    :param image_name: name of the image
    :param dtype: dtype of the numpy array
    :param dsize: size of the numpy array
    """
    descriptors_path = get_descriptors_fullpath(input_path, image_name)
    return image_descriptors_from_file(descriptors_path, dtype, dsize)


def compute_matches(input_path: str,
                    pairsfile_path: str):
    """
    compute matches from descriptors. images to match are selected from a pairsfile (csv with name1, name2, score)

    :param input_path: input path to kapture input root directory
    :type input_path: str
    :param pairsfile_path: path to pairs file (csv with 3 fields, name1, name2, score)
    :type pairsfile_path: str
    """
    logger.info(f'compute_matches. loading input: {input_path}')
    kdata = kapture_from_dir(input_path)
    assert kdata.sensors is not None
    assert kdata.records_camera is not None
    assert kdata.descriptors is not None

    image_pairs = get_pairs_from_file(pairsfile_path)
    matcher = MatchPairNnTorch(use_cuda=torch.cuda.is_available())
    new_matches = kapture.Matches()

    logger.info(f'compute_matches. entering main loop...')
    hide_progress_bar = logger.getEffectiveLevel() > logging.INFO
    for image_path1, image_path2 in tqdm(image_pairs, disable=hide_progress_bar):
        if image_path1 == image_path2:
            continue
        if image_path1 > image_path2:
            image_path1, image_path2 = image_path2, image_path1

        if image_path1 not in kdata.descriptors or image_path2 not in kdata.descriptors:
            logger.warning('unable to find descriptors for image pair : '
                           '\n\t{} \n\t{}'.format(image_path1, image_path2))
            continue

        descriptor1 = load_descriptors(input_path, image_path1, kdata.descriptors.dtype, kdata.descriptors.dsize)
        descriptor2 = load_descriptors(input_path, image_path2, kdata.descriptors.dtype, kdata.descriptors.dsize)
        matches = matcher.match_descriptors(descriptor1, descriptor2)
        matches_path = get_matches_fullpath((image_path1, image_path2), input_path)
        image_matches_to_file(matches_path, matches)
        new_matches.add(image_path1, image_path2)

    if not matches_check_dir(new_matches, input_path):
        logger.critical('matching ended successfully but not all files were saved')
    logger.info('all done')


def compute_matches_command_line():
    parser = argparse.ArgumentParser(
        description='Compute matches with nearest neighbors from descriptors.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
                                  action=kapture.utils.logging.VerbosityParser,
                                  help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument('-q', '--silent', '--quiet',
                                  action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-i', '--input', required=True,
                        help=('input path to kapture input root directory\n'
                              'it must contain all images (query + train) and their local features'))
    parser.add_argument('--pairsfile-path',
                        required=True,
                        type=str,
                        help=('text file in the csv format; where each line is image_name1, image_name2, score '
                              'which contains the image pairs to match'))

    args = parser.parse_args()
    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    logger.debug(''.join(['\n\t{:13} = {}'.format(k, v)
                          for k, v in vars(args).items()]))
    compute_matches(args.input, args.pairsfile_path)


if __name__ == '__main__':
    compute_matches_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Kapture to openmvg export functions.
"""

import json
import logging
import os
import os.path as path
import pathlib
import shutil
from typing import Dict, List, Union
import quaternion
# kapture
import kapture
import kapture.io.csv as kcsv
from kapture.io.records import TransferAction, get_image_fullpath
import kapture.io.structure
from kapture.core.Trajectories import rigs_remove_inplace
from kapture.utils.paths import safe_remove_file, safe_remove_any_path
# local
from .openmvg_commons import DEFAULT_JSON_FILE_NAME, SFM_DATA_VERSION, SFM_DATA_VERSION_NUMBER, ROOT_PATH, INTRINSICS,\
    VIEWS, VIEW_PRIORS, EXTRINSICS, KEY, VALUE, POLYMORPHIC_ID, PTR_WRAPPER, ID, DATA, LOCAL_PATH, FILENAME, ID_VIEW,\
    ID_INTRINSIC, ID_POSE, POLYMORPHIC_NAME, VALUE0, WIDTH, HEIGHT, FOCAL_LENGTH, PRINCIPAL_POINT,\
    DISTO_K1, DISTO_K3, DISTO_T2, FISHEYE, USE_POSE_CENTER_PRIOR, CENTER_WEIGHT, CENTER, USE_POSE_ROTATION_PRIOR,\
    ROTATION_WEIGHT, ROTATION, STRUCTURE, CONTROL_POINTS
from .openmvg_commons import CameraModel


logger = logging.getLogger('openmvg')  # Using global openmvg logger

NEW_ID_MASK = 1 << 31                  # 10000000 00000000 00000000 00000000
VIEW_SPECIAL_POLYMORPHIC_ID = 1 << 30  # 01000000 00000000 00000000 00000000
DEFAULT_FOCAL_LENGTH_FACTOR = 1.2


def _get_data(camera_params: list) -> Dict:
    # w, h, f, cx, cy
    data: Dict[str, Union[int, float, List[float]]] = {WIDTH: int(camera_params[0]),
                                                       HEIGHT: int(camera_params[1]),
                                                       FOCAL_LENGTH: float(camera_params[2]),
                                                       PRINCIPAL_POINT: [float(camera_params[3]),
                                                                         float(camera_params[4])]}
    return data


def _get_intrinsic_pinhole(camera_params: list) -> Dict:
    # w, h, f, cx, cy
    return _get_data(camera_params)


def _get_intrinsic_pinhole_radial_k1(camera_params: list) -> Dict:
    data = _get_data(camera_params)
    # w, h, f, cx, cy, k
    data[DISTO_K1] = [float(camera_params[5])]
    return data


def _get_intrinsic_pinhole_radial_k3(camera_params: list) -> Dict:
    data = _get_data(camera_params)
    # w, h, f, cx, cy, k1, k2, k3
    data[DISTO_K3] = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7])]
    return data


def _get_intrinsic_pinhole_brown_t2(camera_params: list) -> Dict:
    # w, h, f, cx, cy, k1, k2, k3, t1, t2
    disto_t2 = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7]),
                float(camera_params[8]), float(camera_params[9])]
    return {VALUE0: _get_data(camera_params), DISTO_T2: disto_t2}


def _get_intrinsic_fisheye(camera_params: list) -> Dict:
    # w, h, f, cx, cy, k1, k2, k3, k4
    fisheye = [float(camera_params[5]), float(camera_params[6]), float(camera_params[7]), float(camera_params[8])]
    return {VALUE0: _get_data(camera_params),
            FISHEYE: fisheye}


def load_kapture(kapture_path: str) -> kapture.Kapture:
    """
    Load the kapture data stored at the given path

    :param kapture_path: kapture data top directory
    :return: a kapture object
    """
    logger.info(f'Loading kapture data from {kapture_path}...')
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_path)
    logger.info('loaded')
    assert isinstance(kapture_data, kapture.Kapture)
    return kapture_data


def kapture_to_openmvg(kapture_data: kapture.Kapture, kapture_path: str,
                       image_action: TransferAction, openmvg_path: str) -> Dict:
    """
    Convert the kapture data into an openMVG dataset stored as a dictionary.
    The format is defined here:
    https://openmvg.readthedocs.io/en/latest/software/SfM/SfM_OutputFormat/

    :param kapture_data: the kapture data
    :param kapture_path: top directory of the kapture data and the images
    :param image_action: action to apply on images: link, copy, move or do nothing.
    :param openmvg_path: top directory of the openmvg data and images
    :return: an SfM_data, the openmvg structure, stored as a dictionary ready to be serialized
    """

    assert kapture_data.cameras is not None
    assert kapture_data.records_camera is not None
    cameras = kapture_data.cameras
    # Check we don't have other sensors defined
    extra_sensor_number = len(kapture_data.sensors)-len(cameras)
    if extra_sensor_number > 0:
        logger.warning(f'We will ignore {extra_sensor_number} sensors that are not camera')
    records_camera = kapture_data.records_camera
    all_records_camera = list(kapture.flatten(records_camera))
    trajectories = kapture_data.trajectories

    # openmvg does not support rigs
    if kapture_data.rigs:
        logger.info('remove rigs notation.')
        rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)
        kapture_data.rigs.clear()

    # Compute root path and camera used in records
    sub_root_path: str = ''
    image_dirs = {}
    used_cameras = {}
    for _, cam_id, name in all_records_camera:
        used_cameras[cam_id] = cam_id
        img_dir = path.dirname(name)
        image_dirs[img_dir] = img_dir
    if len(image_dirs) > 1:
        # Find if they share a top path
        image_dirs_list = list(image_dirs.keys())
        sub_root_path = path.commonpath(image_dirs_list)
    elif len(image_dirs) == 1:
        sub_root_path = next(iter(image_dirs.keys()))
    if image_action == TransferAction.skip:
        root_path = kapture_path
    else:  # We will create a new hierarchy of images
        root_path = openmvg_path
    root_path = os.path.abspath(path.join(root_path, sub_root_path))
    if image_action == TransferAction.root_link:
        if not sub_root_path:
            # We can not link directly to the top destination openmvg directory
            # We need an additional level
            root_path = path.join(root_path, 'images')
        kapture_records_path = get_image_fullpath(kapture_path, sub_root_path)
        # Do a unique images directory link
        # openmvg_root_path -> kapture/<records_dir>/openmvg_top_images_directory
        # beware that the paths are reverted in the symlink call
        os.symlink(kapture_records_path, root_path)
    sfm_data = {SFM_DATA_VERSION: SFM_DATA_VERSION_NUMBER,
                ROOT_PATH: root_path}

    views = []
    intrinsics = []
    extrinsics = []
    polymorphic_id_current = 1
    ptr_wrapper_id_current = 1
    polymorphic_id_types = {}

    # process all cameras
    for cam_id, camera in cameras.items():
        # Ignore not used cameras
        if not used_cameras.get(cam_id):
            logger.warning(f'Skipping camera definition {cam_id} {camera.name} without recorded images.')
            continue
        cam_type = camera.camera_type
        camera_params = camera.camera_params
        if cam_type == kapture.CameraType.SIMPLE_PINHOLE:
            # w, h, f, cx, cy
            model_used = CameraModel.pinhole
            data = _get_intrinsic_pinhole(camera_params)
        elif cam_type == kapture.CameraType.PINHOLE:
            # w, h, f, cx, cy
            model_used = CameraModel.pinhole
            faked_params = [camera_params[0], camera_params[1],  # width height
                            (camera_params[2] + camera_params[3])/2,  # fx+fy/2 as f
                            camera_params[4], camera_params[5]]  # cx cy
            data = _get_intrinsic_pinhole(faked_params)
        elif cam_type == kapture.CameraType.SIMPLE_RADIAL:
            # w, h, f, cx, cy, k
            model_used = CameraModel.pinhole_radial_k1
            data = _get_intrinsic_pinhole_radial_k1(camera_params)
        elif cam_type == kapture.CameraType.RADIAL:
            # w, h, f, cx, cy, k1, k2, k3
            model_used = CameraModel.pinhole_radial_k3
            faked_params = [camera_params[0], camera_params[1],  # width height
                            camera_params[2],  # f
                            camera_params[3], camera_params[4],  # cx cy
                            camera_params[5], camera_params[6], 0  # k1, k2, k3
                            ]
            data = _get_intrinsic_pinhole_radial_k3(faked_params)
        elif cam_type == kapture.CameraType.FULL_OPENCV or cam_type == kapture.CameraType.OPENCV:
            # w, h, f, cx, cy, k1, k2, k3, t1, t2
            model_used = CameraModel.pinhole_brown_t2
            k3 = camera_params[10] if len(camera_params) > 10 else 0
            faked_params = [camera_params[0], camera_params[1],  # width height
                            (camera_params[2] + camera_params[3])/2,  # fx+fy/2 as f
                            camera_params[4], camera_params[5],  # cx cy
                            camera_params[6], camera_params[7], k3,  # k1, k2, k3
                            camera_params[8], camera_params[9]  # p1, p2 (=t1, t2)
                            ]
            data = _get_intrinsic_pinhole_brown_t2(faked_params)
        elif cam_type == kapture.CameraType.OPENCV_FISHEYE:
            logger.warning('OpenCV fisheye model is not compatible with OpenMVG. Forcing distortion to 0')
            # w, h, f, cx, cy, k1, k2, k3, k4
            model_used = CameraModel.fisheye
            faked_params = [camera_params[0], camera_params[1],  # width height
                            (camera_params[2] + camera_params[3]) / 2,  # fx+fy/2 as f
                            camera_params[4], camera_params[5],  # cx cy
                            0, 0,  # k1, k2
                            0, 0  # k3, k4
                            ]
            data = _get_intrinsic_fisheye(faked_params)
        elif cam_type == kapture.CameraType.RADIAL_FISHEYE or cam_type == kapture.CameraType.SIMPLE_RADIAL_FISHEYE:
            logger.warning('OpenCV fisheye model is not compatible with OpenMVG. Forcing distortion to 0')
            # w, h, f, cx, cy, k1, k2, k3, k4
            model_used = CameraModel.fisheye
            faked_params = [camera_params[0], camera_params[1],  # width height
                            camera_params[2],  # f
                            camera_params[3], camera_params[4],  # cx cy
                            0, 0,  # k1, k2
                            0, 0  # k3, k4
                            ]
            data = _get_intrinsic_fisheye(faked_params)
        elif cam_type == kapture.CameraType.UNKNOWN_CAMERA:
            logger.info(f'Camera {cam_id}: Unknown camera model, using simple radial')
            # Choose simple radial model, to allow openMVG to determine distortion param
            # w, h, f, cx, cy, k
            model_used = CameraModel.pinhole_radial_k1
            faked_params = [camera_params[0], camera_params[1],  # width height
                            max(camera_params[0], camera_params[1])*DEFAULT_FOCAL_LENGTH_FACTOR,  # max(w,h)*1.2 as f
                            int(camera_params[0]/2), int(camera_params[1]/2),  # cx cy
                            0.0]  # k1
            data = _get_intrinsic_pinhole_radial_k1(faked_params)
        else:
            raise ValueError(f'Camera model {cam_type.value} not supported')

        intrinsic = {}
        if model_used not in polymorphic_id_types:
            # if this is the first time model_used is encountered
            # set the first bit of polymorphic_id_current to 1
            intrinsic[POLYMORPHIC_ID] = polymorphic_id_current | NEW_ID_MASK
            intrinsic[POLYMORPHIC_NAME] = model_used.name
            polymorphic_id_types[model_used] = polymorphic_id_current
            polymorphic_id_current += 1
        else:
            intrinsic[POLYMORPHIC_ID] = polymorphic_id_types[model_used]

        # it is assumed that this camera is only encountered once
        # set the first bit of ptr_wrapper_id_current to 1
        data_wrapper = {ID: ptr_wrapper_id_current | NEW_ID_MASK,
                        DATA: data}
        ptr_wrapper_id_current += 1

        intrinsic[PTR_WRAPPER] = data_wrapper
        intrinsics.append({KEY: cam_id, VALUE: intrinsic})

    global_timestamp = 0

    # process all images
    for timestamp, cam_id, kapture_name in all_records_camera:
        local_path = path.dirname(path.relpath(kapture_name, sub_root_path))
        filename = path.basename(kapture_name)
        if image_action != TransferAction.skip and image_action != TransferAction.root_link:
            # Process the image action
            src_path = get_image_fullpath(kapture_path, kapture_name)
            dst_path = path.join(root_path, local_path, filename)
            dst_dir = path.dirname(dst_path)
            if not path.isdir(dst_dir):
                os.makedirs(dst_dir, exist_ok=True)
            # Check if already exist
            if path.exists(dst_path):
                os.unlink(dst_path)
            # Create file or link
            if image_action == TransferAction.copy:
                shutil.copy2(src_path, dst_path)
            elif image_action == TransferAction.move:
                shutil.move(src_path, dst_path)
            else:
                # Link
                if image_action == TransferAction.link_relative:
                    # Compute relative path
                    src_path = path.relpath(src_path, dst_dir)
                os.symlink(src_path, dst_path)

        camera_params = cameras[cam_id].camera_params
        view_data = {LOCAL_PATH: local_path,
                     FILENAME: filename,
                     WIDTH: int(camera_params[0]),
                     HEIGHT: int(camera_params[1]),
                     ID_VIEW: global_timestamp,
                     ID_INTRINSIC: cam_id,
                     ID_POSE: global_timestamp}

        view = {}
        # retrieve image pose from trajectories
        if timestamp not in trajectories:
            view[POLYMORPHIC_ID] = VIEW_SPECIAL_POLYMORPHIC_ID
        else:
            # there is a pose for that timestamp
            # The poses are stored both as priors (in the 'views' table) and as known poses (in the 'extrinsics' table)
            assert cam_id in trajectories[timestamp]
            if VIEW_PRIORS not in polymorphic_id_types:
                # if this is the first time view_priors is encountered
                # set the first bit of polymorphic_id_current to 1
                view[POLYMORPHIC_ID] = polymorphic_id_current | NEW_ID_MASK
                view[POLYMORPHIC_NAME] = VIEW_PRIORS
                polymorphic_id_types[VIEW_PRIORS] = polymorphic_id_current
                polymorphic_id_current += 1
            else:
                view[POLYMORPHIC_ID] = polymorphic_id_types[VIEW_PRIORS]

            pose_tr = trajectories[timestamp].get(cam_id)
            prior_q = pose_tr.r
            prior_t = pose_tr.inverse().t_raw
            pose_data = {CENTER: prior_t,
                         ROTATION: quaternion.as_rotation_matrix(prior_q).tolist()}

            view_data[USE_POSE_CENTER_PRIOR] = True
            view_data[CENTER_WEIGHT] = [1.0, 1.0, 1.0]
            view_data[CENTER] = prior_t
            view_data[USE_POSE_ROTATION_PRIOR] = True
            view_data[ROTATION_WEIGHT] = 1.0
            view_data[ROTATION] = pose_data[ROTATION]
            extrinsics.append({KEY: global_timestamp, VALUE: pose_data})

        # it is assumed that this view is only encountered once
        # set the first bit of ptr_wrapper_id_current to 1
        view_wrapper = {ID: ptr_wrapper_id_current | NEW_ID_MASK,
                        DATA: view_data}
        ptr_wrapper_id_current += 1

        view[PTR_WRAPPER] = view_wrapper
        views.append({KEY: global_timestamp, VALUE: view})

        global_timestamp += 1

    sfm_data[VIEWS] = views
    sfm_data[INTRINSICS] = intrinsics
    sfm_data[EXTRINSICS] = extrinsics
    sfm_data[STRUCTURE] = []
    sfm_data[CONTROL_POINTS] = []

    return sfm_data


def export_openmvg(kapture_path: str, openmvg_path: str,
                   image_action: TransferAction, force: bool = False) -> None:
    """
    Export the kapture data to an openMVG JSON file.
    If the openmvg_path is a directory, it will create a JSON file (using the default name sfm_data.json)
    in that directory.

    :param kapture_path: full path to the top kapture directory
    :param openmvg_path: path of the file or directory where to store the data as JSON
    :param image_action: an action to apply on the images: relative linking, absolute linking, copy or move. Or top
     directory linking or skip to do nothing.
    :param force: if true, will remove existing openMVG data without prompting the user.
    """

    if path.isdir(openmvg_path):  # Existing directory
        json_file = path.join(openmvg_path, DEFAULT_JSON_FILE_NAME)
    else:
        file_ext = path.splitext(openmvg_path)[1]
        if len(file_ext) == 0:  # No extension: -> new directory
            json_file = path.join(openmvg_path, DEFAULT_JSON_FILE_NAME)
        elif file_ext.lower() != '.json':
            logger.warning(f'Creating output directory with file extension {file_ext}')
            json_file = path.join(openmvg_path, DEFAULT_JSON_FILE_NAME)
        else:  # Json file
            json_file = openmvg_path
    json_dir = path.dirname(json_file)
    safe_remove_file(json_file, force)
    if path.exists(json_file):
        raise ValueError(f'{json_file} file already exist')
    if image_action != TransferAction.skip and path.exists(json_dir) and any(pathlib.Path(json_dir).iterdir()):
        safe_remove_any_path(json_dir, force)
        if path.isdir(json_dir):
            raise ValueError(f'Images directory {json_dir} exist with remaining files')
    os.makedirs(json_dir, exist_ok=True)

    kapture_data = load_kapture(kapture_path)
    openmvg_data = kapture_to_openmvg(kapture_data, kapture_path, image_action, json_dir)
    logger.info(f'Saving to openmvg file {json_file}')
    with open(json_file, "w") as fid:
        json.dump(openmvg_data, fid, indent=4)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Openmvg to kapture import functions.
"""

import json
import logging
import quaternion
import numpy as np
import os
import os.path as path
import shutil
from tqdm import tqdm
from typing import Dict, Union
# kapture
import kapture
import kapture.io.csv as kcsv
import kapture.io.structure
from kapture.io.records import TransferAction, get_image_fullpath
from kapture.utils.paths import path_secure
# local
from .openmvg_commons import DEFAULT_JSON_FILE_NAME, SFM_DATA_VERSION, ROOT_PATH, INTRINSICS, VIEWS, EXTRINSICS,\
    KEY, VALUE, POLYMORPHIC_ID, PTR_WRAPPER, ID, DATA, LOCAL_PATH, FILENAME, ID_VIEW, ID_INTRINSIC, ID_INTRINSIC,\
    ID_POSE, POLYMORPHIC_NAME, VALUE0, WIDTH, HEIGHT, FOCAL_LENGTH, PRINCIPAL_POINT, DISTO_K1, DISTO_K3, DISTO_T2, \
    ROTATION, CENTER
from .openmvg_commons import CameraModel

logger = logging.getLogger('openmvg')  # Using global openmvg logger


def import_openmvg(input_path: str,
                   kapture_path: str,
                   image_action: TransferAction,
                   force_overwrite_existing: bool = False) -> None:
    """
    Converts an openMVG JSON file to a kapture directory.
    If an image action is provided (link, copy or move), links to the image files are created,
    or the image files are copied or moved.

    :param input_path: path to the openMVG file (or its directory)
    :param kapture_path: path to the kapture directory where the data will be exported
    :param image_action: action to apply to the images
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    """

    os.makedirs(kapture_path, exist_ok=True)
    kapture.io.structure.delete_existing_kapture_files(kapture_path, force_overwrite_existing)
    if path.isdir(input_path):
        input_file = path.join(input_path, DEFAULT_JSON_FILE_NAME)
    else:
        input_file = input_path
    if not path.isfile(input_file):
        raise ValueError(f'OpenMVG JSON file {input_file} does not exist')

    logger.info(f'Loading openmvg file {input_file}')

    with open(input_file, 'r') as f:
        input_json = json.load(f)
        kapture_data = openmvg_to_kapture(input_json, kapture_path, image_action)
        logger.info(f'Saving to kapture {kapture_path}')
        kcsv.kapture_to_dir(kapture_path, kapture_data)


GET_ID_MASK = 2147483647  # 01111111 11111111 11111111 11111111
ID_POSE_NOT_LOCALIZED = 4294967295  # 11111111 11111111 11111111 11111111


def openmvg_to_kapture(input_json: Dict[str, Union[str, Dict]],
                       kapture_images_path: str,
                       image_action=TransferAction.skip) -> kapture.Kapture:
    """
    Convert an openMVG structure to a kapture object. Also copy, move or link the images files if necessary.

    :param input_json: the openmvg JSON parsed as a dictionary
    :param kapture_images_path: top directory to create the kapture images tree
    :param image_action: action to apply on images: link, copy, move or do nothing.
    :return: the constructed kapture object
    """

    polymorphic_id_to_value = {}
    root_path: str = ''

    if input_json[ROOT_PATH]:
        root_path = input_json[ROOT_PATH]
    elif image_action == TransferAction.skip:
        logger.warning("No root_path in input file")
    else:  # It is needed to execute an action with the image file
        raise ValueError(f"Missing root_path to do image action '{image_action.name}'")
    openmvg_images_dir = path.basename(root_path)

    kapture_cameras = kapture.Sensors()
    if input_json.get(INTRINSICS):
        logger.info(f'Importing intrinsics')
        for sensor in input_json[INTRINSICS]:
            value = sensor[VALUE]
            if POLYMORPHIC_NAME in value:
                # new type name: store it for next instances
                polymorphic_id = value[POLYMORPHIC_ID] & GET_ID_MASK
                polymorphic_id_to_value[polymorphic_id] = value[POLYMORPHIC_NAME]
                logger.debug("New camera_type: "+polymorphic_id_to_value[polymorphic_id])
            else:
                if POLYMORPHIC_ID not in value:
                    raise ValueError(f'{POLYMORPHIC_ID} is missing (intrinsics)')
                polymorphic_id = value[POLYMORPHIC_ID]

            if polymorphic_id not in polymorphic_id_to_value:
                raise ValueError(f'Unknown polymorphic_id {polymorphic_id}')

            camera_model = CameraModel(polymorphic_id_to_value[polymorphic_id])
            camera_data = value[PTR_WRAPPER][DATA]

            if camera_model == CameraModel.pinhole:
                # w, h, f, cx, cy
                camera = kapture.Camera(kapture.CameraType.SIMPLE_PINHOLE, [
                    int(camera_data[WIDTH]),
                    int(camera_data[HEIGHT]),
                    camera_data[FOCAL_LENGTH],
                    camera_data[PRINCIPAL_POINT][0],
                    camera_data[PRINCIPAL_POINT][1],
                ])
            elif camera_model == CameraModel.pinhole_radial_k1:
                # w, h, f, cx, cy, k
                camera = kapture.Camera(kapture.CameraType.SIMPLE_RADIAL, [
                    int(camera_data[WIDTH]),
                    int(camera_data[HEIGHT]),
                    camera_data[FOCAL_LENGTH],
                    camera_data[PRINCIPAL_POINT][0],
                    camera_data[PRINCIPAL_POINT][1],
                    camera_data[DISTO_K1][0]
                ])
            elif camera_model == CameraModel.pinhole_radial_k3:
                # w, h, f, cx, cy, k1, k2, k3
                camera = kapture.Camera(kapture.CameraType.RADIAL, [
                    int(camera_data[WIDTH]),
                    int(camera_data[HEIGHT]),
                    camera_data[FOCAL_LENGTH],
                    camera_data[PRINCIPAL_POINT][0],
                    camera_data[PRINCIPAL_POINT][1],
                    camera_data[DISTO_K3][0],
                    camera_data[DISTO_K3][1]
                ])
                # camera_data["disto_k3"][2] ignored: radial model has two distortion param, while openMVG's has three
            elif camera_model == CameraModel.pinhole_brown_t2:
                # w, h, f, cx, cy, k1, k2, k3, t1, t2
                if float(camera_data[DISTO_T2][2]) != 0:
                    # if k3 not null, use FULL_OPENCV, otherwise OPENCV
                    # w, h, fx, fy, cx, cy, k1, k2, p1, p2, k3, k4, k5, k6
                    value0 = camera_data[VALUE0]
                    disto_t2 = camera_data[DISTO_T2]
                    camera = kapture.Camera(kapture.CameraType.FULL_OPENCV, [
                        int(value0[WIDTH]),
                        int(value0[HEIGHT]),
                        value0[FOCAL_LENGTH],
                        value0[FOCAL_LENGTH],
                        value0[PRINCIPAL_POINT][0],
                        value0[PRINCIPAL_POINT][1],
                        disto_t2[0], disto_t2[1], disto_t2[3], disto_t2[4], disto_t2[2],
                        0, 0, 0
                    ])
                else:
                    # w, h, fx, fy, cx, cy, k1, k2, p1, p2
                    value0 = camera_data[VALUE0]
                    disto_t2 = camera_data[DISTO_T2]
                    camera = kapture.Camera(kapture.CameraType.OPENCV, [
                        int(value0[WIDTH]),
                        int(value0[HEIGHT]),
                        value0[FOCAL_LENGTH],
                        value0[FOCAL_LENGTH],
                        value0[PRINCIPAL_POINT][0],
                        value0[PRINCIPAL_POINT][1],
                        disto_t2[0], disto_t2[1], disto_t2[3], disto_t2[4]])
            elif camera_model == CameraModel.fisheye:
                logger.warning(
                    "OpenMVG fisheye models are not compatible with OpenCV."
                    " Using SIMPLE_RADIAL_FISHEYE and forcing distortion to 0")
                # w, h, f, cx, cy, k
                value0 = camera_data[VALUE0]
                camera = kapture.Camera(kapture.CameraType.SIMPLE_RADIAL_FISHEYE, [
                    int(value0[WIDTH]),
                    int(value0[HEIGHT]),
                    value0[FOCAL_LENGTH],
                    value0[PRINCIPAL_POINT][0],
                    value0[PRINCIPAL_POINT][1],
                    0])
            else:
                raise ValueError(f'Camera model {camera_model} not supported')

            kapture_cameras[str(sensor[KEY])] = camera

    device_identifiers = {int: str}  # Pose id -> device id
    timestamp_for_pose = {int: int}  # Pose id -> timestamp
    records_camera = kapture.RecordsCamera()
    if input_json.get(VIEWS):
        views = input_json[VIEWS]
        if image_action == TransferAction.root_link:
            # Do a unique images directory link
            # kapture/<records_dir>/openmvg_top_images_directory -> openmvg_root_path
            kapture_records_path = get_image_fullpath(kapture_images_path)
            os.makedirs(kapture_records_path, exist_ok=True)
            os.symlink(root_path, path.join(kapture_records_path, openmvg_images_dir))
        logger.info(f'Importing {len(views)} images')
        # Progress bar only in debug or info level
        if image_action != TransferAction.skip and image_action != TransferAction.root_link\
                and logger.getEffectiveLevel() <= logging.INFO:
            progress_bar = tqdm(total=len(views))
        else:
            progress_bar = None
        for view in views:
            input_data = view[VALUE][PTR_WRAPPER][DATA]
            pose_id = input_data[ID_POSE]
            # All two values should be the same (?)
            if input_data[ID_VIEW]:
                timestamp = input_data[ID_VIEW]
            else:
                timestamp = view[KEY]
            device_id = str(input_data[ID_INTRINSIC])  # device_id must be a string for kapture
            device_identifiers[pose_id] = device_id
            timestamp_for_pose[pose_id] = timestamp
            filename: str
            if input_data.get(LOCAL_PATH):
                filename = path.join(input_data[LOCAL_PATH], input_data[FILENAME])
            else:
                filename = input_data[FILENAME]
            if root_path:
                src_path = path.join(root_path, filename)
            else:
                src_path = filename

            # Add the common openmvg images directory in front of the filename
            kapture_filename = path.join(openmvg_images_dir, filename)
            if image_action != TransferAction.skip and image_action != TransferAction.root_link:
                dst_path = get_image_fullpath(kapture_images_path, kapture_filename)
                # Create destination directory if necessary
                dst_dir = path.dirname(dst_path)
                if not path.isdir(dst_dir):
                    os.makedirs(dst_dir, exist_ok=True)
                # Check if already exist
                if path.exists(dst_path):
                    os.unlink(dst_path)
                # Create file or link
                if image_action == TransferAction.copy:
                    shutil.copy2(src_path, dst_path)
                elif image_action == TransferAction.move:
                    shutil.move(src_path, dst_path)
                else:
                    # Individual link
                    if image_action == TransferAction.link_relative:
                        # Compute relative path
                        src_path = path.relpath(src_path, dst_dir)
                    os.symlink(src_path, dst_path)
                    # This might crash on Windows if the user executing this code has no admin privilege
                progress_bar and progress_bar.update(1)

            key = (timestamp, device_id)  # tuple of int,str
            records_camera[key] = path_secure(kapture_filename)
        progress_bar and progress_bar.close()

    trajectories = kapture.Trajectories()
    if input_json.get(EXTRINSICS):
        extrinsics = input_json[EXTRINSICS]
        logger.info(f'Importing {len(extrinsics)} extrinsics -> trajectories')
        for pose in extrinsics:
            pose_id = pose[KEY]
            center = pose[VALUE][CENTER]
            rotation = pose[VALUE][ROTATION]
            kap_translation = -1 * np.matmul(rotation, center)
            kap_pose = kapture.PoseTransform(quaternion.from_rotation_matrix(rotation), kap_translation)
            timestamp = timestamp_for_pose.get(pose_id)
            if timestamp is None:
                logger.warning(f'Missing timestamp for extrinsic {pose_id}')
                continue
            device_id = device_identifiers.get(pose_id)
            if device_id is None:
                logger.warning(f'Missing device for extrinsic {pose_id}')
                continue
            trajectories[(timestamp, device_id)] = kap_pose  # tuple of int,str -> 6D pose

    kapture_data = kapture.Kapture(sensors=kapture_cameras, records_camera=records_camera, trajectories=trajectories)
    return kapture_data



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license
from abc import ABC, abstractmethod
import torch
import numpy as np
from kapture.utils.logging import getLogger


class MatchPairGenerator(ABC):
    @abstractmethod
    def match_descriptors(self, descriptors_1, descriptors_2):
        raise NotImplementedError()


class MatchPairNnTorch(MatchPairGenerator):
    def __init__(self, use_cuda=True):
        super().__init__()
        self._device = torch.device("cuda:0"
                                    if use_cuda and torch.cuda.is_available()
                                    else "cpu")

    def match_descriptors(self, descriptors_1, descriptors_2):
        if descriptors_1.shape[0] == 0 or descriptors_2.shape[0] == 0:
            return np.zeros((0, 3))

        # send data to GPU
        descriptors1_torch = torch.from_numpy(descriptors_1).to(self._device)
        descriptors2_torch = torch.from_numpy(descriptors_2).to(self._device)
        # make sure its double (because CUDA tensors only supports floating-point)
        descriptors1_torch = descriptors1_torch.float()
        descriptors2_torch = descriptors2_torch.float()
        # sanity check
        if not descriptors1_torch.device == self._device:
            getLogger().debug('descriptor on device {} (requested {})'.format(descriptors1_torch.device, self._device))
        if not descriptors2_torch.device == self._device:
            getLogger().debug('descriptor on device {} (requested {})'.format(descriptors2_torch.device, self._device))

        simmilarity_matrix = descriptors1_torch @ descriptors2_torch.t()
        scores = torch.max(simmilarity_matrix, dim=1)[0]
        nearest_neighbor_idx_1vs2 = torch.max(simmilarity_matrix, dim=1)[1]
        nearest_neighbor_idx_2vs1 = torch.max(simmilarity_matrix, dim=0)[1]
        ids1 = torch.arange(0, simmilarity_matrix.shape[0], device=descriptors1_torch.device)
        # cross check
        mask = ids1 == nearest_neighbor_idx_2vs1[nearest_neighbor_idx_1vs2]
        matches_torch = torch.stack(
            [ids1[mask].type(torch.float), nearest_neighbor_idx_1vs2[mask].type(torch.float), scores[mask]]).t()
        # retrieve data back from GPU
        matches = matches_torch.data.cpu().numpy()
        matches = matches.astype(np.float)
        return matches



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script exports data from kapture to COLMAP database and/or reconstruction files.
"""

import logging
import os
import os.path as path
from typing import Optional
import warnings

# kapture
import kapture
from kapture.core.Trajectories import rigs_remove_inplace
import kapture.io.csv as csv
import kapture.io.structure
import kapture.algo.merge_keep_ids

# local
from .database import COLMAPDatabase
from .database_extra import is_colmap_db_empty, kapture_to_colmap, get_colmap_image_ids_from_db,\
    get_colmap_camera_ids_from_db
from .export_colmap_rigs import export_colmap_rig
from .export_colmap_reconstruction import export_to_colmap_txt

logger = logging.getLogger('colmap')


def export_colmap(kapture_dirpath: str,
                  colmap_database_filepath: str,
                  colmap_reconstruction_dirpath: Optional[str],
                  colmap_rig_filepath: str = None,
                  force_overwrite_existing: bool = False) -> None:
    """
    Exports kapture data to colmap database and or reconstruction text files.

    :param kapture_dirpath: kapture top directory
    :param colmap_database_filepath: path to colmap database file
    :param colmap_reconstruction_dirpath: path to colmap reconstruction directory
    :param colmap_rig_filepath: path to colmap rig file
    :param force_overwrite_existing: Silently overwrite colmap files if already exists.
    """

    os.makedirs(path.dirname(colmap_database_filepath), exist_ok=True)
    if colmap_reconstruction_dirpath:
        os.makedirs(colmap_reconstruction_dirpath, exist_ok=True)

    assert colmap_database_filepath
    if path.isfile(colmap_database_filepath):
        to_delete = force_overwrite_existing or (input(
            'database file already exist, would you like to delete it ? [y/N]').lower() == 'y')
        if to_delete:
            logger.info('deleting already existing {}'.format(colmap_database_filepath))
            os.remove(colmap_database_filepath)

    logger.info('creating colmap database in {}'.format(colmap_database_filepath))
    db = COLMAPDatabase.connect(colmap_database_filepath)
    if not is_colmap_db_empty(db):
        raise ValueError('the existing colmap database is not empty : {}'.format(colmap_database_filepath))

    logger.info('loading kapture files...')
    kapture_data = csv.kapture_from_dir(kapture_dirpath)
    assert isinstance(kapture_data, kapture.Kapture)

    # COLMAP does not fully support rigs.
    if kapture_data.rigs is not None:
        # make sure, rigs are not used in trajectories.
        logger.info('remove rigs notation.')
        rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)

    # write colmap database
    kapture_to_colmap(kapture_data, kapture_dirpath, db)

    if colmap_reconstruction_dirpath:
        # create text files
        colmap_camera_ids = get_colmap_camera_ids_from_db(db, kapture_data.records_camera)
        colmap_image_ids = get_colmap_image_ids_from_db(db)
        export_to_colmap_txt(
            colmap_reconstruction_dirpath=colmap_reconstruction_dirpath,
            kapture_data=kapture_data,
            kapture_dirpath=kapture_dirpath,
            colmap_camera_ids=colmap_camera_ids,
            colmap_image_ids=colmap_image_ids)
        if colmap_rig_filepath:
            try:
                if kapture_data.rigs is None:
                    raise ValueError('No rig to export.')
                export_colmap_rigs(colmap_rig_filepath,
                                   kapture_data.rigs,
                                   kapture_data.records_camera,
                                   colmap_camera_ids)
            except Exception as e:
                warnings.warn(e)
                logger.warning(e)

    db.close()
